"""Thermostat Status Response Handler."""
import logging

from .. import inbound_handler
from ...address import Address
from ...constants import MessageFlagType, ThermostatMode
from ...topics import THERMOSTAT_STATUS_RESPONSE
from ...utils import build_topic, calc_thermostat_mode, calc_thermostat_temp, bit_is_set
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)

SYS_MODE_MAP = {
    0: ThermostatMode.OFF,
    1: ThermostatMode.AUTO,
    2: ThermostatMode.HEAT,
    3: ThermostatMode.COOL,
}


def _parse_status_flag(status_flag):
    """Parse the status flag."""
    cooling = bit_is_set(status_flag, 0)
    heating = bit_is_set(status_flag, 1)
    energy = bit_is_set(status_flag, 2)
    celcius = bit_is_set(status_flag, 3)
    hold = bit_is_set(status_flag, 4)
    return (cooling, heating, energy, celcius, hold)


class ThermostatStatusResponseHandler(InboundHandlerBase):
    """Thermostat Status Response Handler."""

    def __init__(self, address):
        """Init the ThermostatStatusResponseHandler class."""
        self._address = Address(address)
        super().__init__(
            topic=THERMOSTAT_STATUS_RESPONSE,
            address=self._address,
            message_type=MessageFlagType.DIRECT,
        )
        self._subscriber_topic = build_topic(
            prefix="handler.{}".format(self._address.id),  # Force address
            topic=THERMOSTAT_STATUS_RESPONSE,
            message_type=MessageFlagType.DIRECT,
        )

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the Status Response from a Thermostat."""
        day = user_data["d2"]
        hour = user_data["d3"]
        minute = user_data["d4"]
        second = user_data["d5"]
        system_mode, fan_mode = calc_thermostat_mode(
            user_data["d6"], SYS_MODE_MAP, sys_low=False
        )
        cool_set_point = user_data["d7"]
        humidity = user_data["d8"]
        temp = calc_thermostat_temp(user_data["d9"], user_data["d10"])
        cooling, heating, _, celsius, _ = _parse_status_flag(user_data["d11"])
        heat_set_point = user_data["d12"]
        self._call_subscribers(
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            system_mode=system_mode,
            fan_mode=fan_mode,
            cool_set_point=cool_set_point,
            humidity=humidity,
            temperature=temp,
            cooling=cooling,
            heating=heating,
            celsius=celsius,
            heat_set_point=heat_set_point,
        )



"""Modem command to get next ALDB record."""
import logging

from . import ack_handler
from ..topics import GET_IM_INFO
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetImInfoHandler(OutboundHandlerBase):
    """Handle Get IM Info commands."""

    def __init__(self):
        """Init the GetNextAldbRecordNak class."""
        super().__init__(topic=GET_IM_INFO)

    @ack_handler()
    def handle_ack(self, address, cat, subcat, firmware):
        """Receive the ACK message and return True."""
        self._call_subscribers(
            address=address, cat=cat, subcat=subcat, firmware=firmware
        )



"""On Leve state."""
import logging

from .group_base import GroupBase
from ..address import Address

_LOGGER = logging.getLogger(__name__)


class OnLevel(GroupBase):
    """Variable On Level state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: int = None
    ):
        """Init the OnLevel class."""
        super().__init__(name, address, group, default, value_type=int)

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        self.value = int(on_level) if on_level else 0



"""Insteon Modem ALDB Read Manager."""
import asyncio
import logging

from .. import pub
from ..address import Address
from ..aldb.aldb_record import ALDBRecord
from ..constants import ResponseStatus
from ..handlers.all_link_record_response import AllLinkRecordResponseHandler
from ..handlers.get_first_all_link_record import GetFirstAllLinkRecordHandler
from ..handlers.get_next_all_link_record import GetNextAllLinkRecordHandler
from ..topics import ALL_LINK_RECORD_RESPONSE

_LOGGER = logging.getLogger(__name__)


class ImReadManager:
    """Insteon Modem ALDB Read Manager."""

    def __init__(self, aldb):
        """Init the ImReadManager class."""
        self._aldb = aldb
        self._get_first_handler = GetFirstAllLinkRecordHandler()
        self._get_next_handler = GetNextAllLinkRecordHandler()
        mgr = pub.getDefaultTopicMgr()
        topic = mgr.getTopic(ALL_LINK_RECORD_RESPONSE, okIfNone=True)
        if not topic:
            self._receive_record_handler = AllLinkRecordResponseHandler()
            self._receive_record_handler.subscribe(self._receive_record)
        self._retries = 0
        self._load_lock = asyncio.Lock()
        self._last_mem_addr = 0x0FFF

    def load(self):
        """Load the Insteon Modem ALDB."""
        asyncio.ensure_future(self.async_load())

    async def async_load(self):
        """Load the Insteon Modem ALDB."""
        response = False
        retries = 3
        await self._load_lock.acquire()
        while response != ResponseStatus.SUCCESS and retries:
            response = await self._get_first_handler.async_send()
            self._retries -= 1
        if response == ResponseStatus.SUCCESS:
            await self._get_next_record()
        return ResponseStatus.SUCCESS

    def _max_retries(self):
        """Test if max retries reached."""
        return bool(self._retries >= 3)

    def _receive_record(
        self,
        in_use: bool,
        high_water_mark: bool,
        controller: bool,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
        bit5: bool,
        bit4: bool,
    ):
        """Receive a record and load into the ALDB."""
        self._last_mem_addr = self._last_mem_addr - 8
        record = ALDBRecord(
            memory=self._last_mem_addr,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        self._aldb[self._last_mem_addr] = record

    async def _get_next_record(self):
        """Get the next ALDB record."""
        response = ResponseStatus.FAILURE
        retries = 3
        unsent = 50
        while retries and unsent:
            response = await self._get_next_handler.async_send()
            if response == ResponseStatus.SUCCESS:
                retries = 3
                unsent = 50
            if response == ResponseStatus.FAILURE:
                retries -= 1
            if response == ResponseStatus.UNSENT:
                unsent -= 1



"""Utility methods."""
import logging
from enum import Enum, IntEnum
import traceback
from typing import Iterable

from . import pub
from .address import Address
from .constants import (
    HC_LOOKUP,
    RAMP_RATES,
    UC_LOOKUP,
    MessageFlagType,
    ResponseStatus,
    X10Commands,
    ThermostatMode,
    FanSpeedRange,
    FanSpeed,
)
from .protocol.commands import commands

_LOGGER = logging.getLogger(__name__)
_LOGGER_TOPICS = logging.getLogger("pyinsteon.topics")


def housecode_to_byte(housecode: str) -> int:
    """Return the byte value of an X10 housecode."""
    return HC_LOOKUP.get(housecode.lower())


def unitcode_to_byte(unitcode: int) -> int:
    """Return the byte value of an X10 unitcode."""
    return UC_LOOKUP.get(unitcode)


def byte_to_housecode(bytecode: int) -> str:
    """Return an X10 housecode value from a byte value."""
    house_code = list(HC_LOOKUP.keys())[list(HC_LOOKUP.values()).index(bytecode)]
    return house_code.upper()


def byte_to_unitcode(bytecode: int) -> int:
    """Return an X10 unitcode value from a byte value."""
    return list(UC_LOOKUP.keys())[list(UC_LOOKUP.values()).index(bytecode)]


def byte_to_command(bytecode: int) -> int:
    """Return an X10 unitcode value from a byte value."""
    return X10Commands(bytecode)


def parse_x10(raw_x10):
    """Convert an X10 message to a dictionary."""
    housecode, uc_or_cmd = raw_x10_to_bytes(raw_x10)
    return byte_to_housecode(housecode), uc_or_cmd


def byte_to_int(bytecode: bytes) -> int:
    """Return an int from a byte string."""
    return int.from_bytes(bytecode, byteorder="big")


def raw_x10_to_bytes(raw_x10: int) -> int:
    """Return the byte value of a raw X10 command."""
    house_code_byte = raw_x10[0]
    uc_or_cmd_byte = raw_x10[1]
    return house_code_byte, uc_or_cmd_byte


def bit_is_set(bitmask: int, bit: int) -> bool:
    """Return True if a specific bit is set in a bitmask.

    Uses the low bit is 1 and the high bit is 8.
    """
    return bool(bitmask & (1 << bit))


def set_bit(data: int, bit: int, is_on: bool) -> int:
    """Set the value of a bit in a bitmask on or off.

    Uses the low bit is 0 and the high bit is 7.
    """
    if isinstance(data, bytes):
        bitmask = int.from_bytes(data, byteorder="big")
    else:
        bitmask = data
    if is_on:
        return bitmask | (1 << bit)
    return bitmask & (0xFF & ~(1 << bit))


def vars_to_bytes(vals: Iterable) -> bytes:
    """Create a byte string from a set of values."""
    msg = bytearray()
    for val in vals:
        if val is None:
            pass
        elif isinstance(val, int):
            msg.extend(bytes([val]))
        else:
            msg.extend(bytes(val))
    return bytes(msg)


def vars_to_string(vals: Iterable) -> str:
    """Create a byte string from a set of values."""

    output = []
    for fld, val in vals:
        if val is None:
            pass
        elif isinstance(val, (Enum, IntEnum)):
            valstr = str(val)
        elif isinstance(val, int):
            valstr = "0x{0:02x}".format(val)
        elif isinstance(val, bytes):
            valstr = "0x{:s}".format(val.hex())
        else:
            valstr = str(val)
        output.append("{}: {}".format(fld, valstr))
    return ", ".join(output)


def vars_to_repr(vals: Iterable) -> str:
    """Create a byte string from a set of values."""
    output = []
    for fld, val in vals:
        if val is None:
            pass
        elif isinstance(val, (Enum, IntEnum)):
            valstr = repr(val)
        elif isinstance(val, int):
            valstr = "0x{0:02x}".format(val)
        elif isinstance(val, bytes):
            valstr = "0x{:s}".format(val.hex())
        else:
            valstr = repr(val)
        output.append("{}: {}".format(fld, valstr))
    return ", ".join(output)


def test_values_eq(val1, val2) -> bool:
    """Test if val1 eq val2 for template management."""
    if val1 is None or val2 is None:
        return True
    if val1 == val2:
        return True
    return False


def _include_address(prefix, topic, address, message_type):
    """Test if we should include the address in the topic build."""
    if address is None:
        return False

    if isinstance(message_type, str):
        message_type = getattr(MessageFlagType, str(message_type).upper())

    if prefix == "send" and message_type == MessageFlagType.DIRECT:
        return False

    if commands.get(topic) is None:
        return False

    return True


def _include_group(topic, group, message_type):
    if group is None:
        return False

    if not commands.use_group(topic):
        return False

    if message_type in [MessageFlagType.ALL_LINK_CLEANUP]:
        return False

    return True


def build_topic(topic, prefix=None, address=None, group=None, message_type=None):
    """Build a full topic from components."""
    full_topic = ""
    if prefix is not None:
        # Adding the . separator since there must be something after a prefix
        full_topic = "{}.".format(str(prefix))

    if _include_address(prefix, topic, address, message_type):
        addr = address.id if isinstance(address, Address) else address
        full_topic = "{}{}.".format(full_topic, addr)
        if commands.use_group(topic) and group is not None:
            group = group if group else 1
            full_topic = "{}{}.".format(full_topic, group)

    full_topic = "{}{}".format(full_topic, topic)
    if message_type is not None:
        full_topic = "{}.{}".format(full_topic, str(message_type))
    return full_topic


def multiple_status(*args):
    """Return the proper status based on the worst case of all status responses."""
    worst_response = 1
    for response in args:
        if response is None:
            continue
        if int(response) == 0 or worst_response == 0:
            worst_response = 0
        elif int(response) > worst_response:
            worst_response = int(response)
    return ResponseStatus(worst_response)


def ramp_rate_to_seconds(ramp_rate: int):
    """Return the seconds associated with a ramp rate."""
    if int(ramp_rate) not in range(0, 31):
        raise ValueError("Ramp rate must be between 0x00 and 0x1f (31)")

    return RAMP_RATES[int(ramp_rate)]


def seconds_to_ramp_rate(seconds: float):
    """Return the ramp rate asscociated with a number of seconds."""
    if seconds > 480:
        raise ValueError("Ramp rate cannot be more than 480 seconds (8 minutes)")

    max_sec = 999
    min_sec = -1
    ramp_rate = None
    for curr_rr in RAMP_RATES:
        rr_secs = RAMP_RATES[curr_rr]
        if rr_secs >= seconds and ((rr_secs - seconds) < (max_sec - seconds)):
            max_sec = rr_secs
            if (max_sec - seconds) < (seconds - min_sec):
                ramp_rate = curr_rr
        if rr_secs <= seconds and ((seconds - rr_secs) < (seconds - min_sec)):
            min_sec = rr_secs
            if (seconds - min_sec) <= (max_sec - seconds):
                ramp_rate = curr_rr
    return ramp_rate


def log_error(msg, ex, topic=None, kwargs=None):
    """Print an error message when a topic cannot be distributed."""
    _LOGGER.error("An issue occured distributing the following message")
    _LOGGER.error("MSG: %s", msg)
    _LOGGER.error("Topic: %s data: %s", topic, kwargs)
    _LOGGER.error("Error: %s", str(ex))
    _LOGGER_TOPICS.debug(traceback.format_exc())
    if topic is not None:
        topic_mgr = pub.getDefaultTopicMgr()
        topic = topic_mgr.getTopic(topic, okIfNone=True)
        if topic:
            for subscriber in topic.getListeners():
                _LOGGER.error("    Subscriber: %s", subscriber)


def to_celsius(fahrenheit):
    """Convert fahrenheit to celsius."""
    return int(round((fahrenheit - 32) * 5 / 9, 0))


def to_fahrenheit(celsius):
    """Convert celsius to fahrenheit."""
    return int(round(celsius * 9 / 5 + 32, 0))


def calc_thermostat_temp(high_byte, low_byte):
    """Calculate the temperature."""
    return (low_byte | (high_byte << 8)) * 0.1


def calc_thermostat_mode(mode_byte, sys_mode_map=None, sys_low=True):
    """Calculate the system and fan mode."""
    if sys_mode_map is None:
        sys_mode_map = {
            int(ThermostatMode.OFF): ThermostatMode.OFF,
            int(ThermostatMode.AUTO): ThermostatMode.AUTO,
            int(ThermostatMode.HEAT): ThermostatMode.HEAT,
            int(ThermostatMode.COOL): ThermostatMode.COOL,
        }

    mode1 = mode_byte & 0x0F
    mode2 = mode_byte >> 4
    system_mode, fan_mode = (mode1, mode2) if sys_low else (mode2, mode1)
    if fan_mode in (0, 4):
        fan_mode = ThermostatMode.FAN_AUTO
    else:
        fan_mode = ThermostatMode.FAN_ALWAYS_ON
    sys_mode = sys_mode_map.get(system_mode)
    if sys_mode is None:
        sys_mode = ThermostatMode.AUTO
    return sys_mode, fan_mode


def publish_topic(topic, logger=None, **kwargs):
    """Publish a topic and log errors."""
    # Send log message as caller not utils.
    if logger is None:
        logger = logging.getLogger(__name__)
    try:
        pub.sendMessage(topic, **kwargs)
    except pub.ExcHandlerError as exc:
        logger.error("pubsub ExcHandlerError")
        logger.error("Error processing topic: %s", topic)
        logger.error("Error in topic listner: %s", exc.badExcListenerID)
    except pub.SenderMissingReqdMsgDataError as exc:
        logger.error("SenderMissingReqdMsgDataError")
        logger.error(str(exc))
        for listner in pub.getDefaultTopicMgr().getTopic(topic).getListeners():
            logger.error("Topic listener: %s", listner)
    except pub.SenderUnknownMsgDataError as exc:
        logger.error("SenderUnknownMsgDataError")
        logger.error(str(exc))
        for listner in pub.getDefaultTopicMgr().getTopic(topic).getListeners():
            logger.error("Topic listener: %s", listner)


def subscribe_topic(listener, topic_name, logger=None):
    """Subscribe a listener to a topic and log errors."""
    topic_mgr = pub.getDefaultTopicMgr()
    topic = topic_mgr.getOrCreateTopic(topic_name)
    if pub.isSubscribed(listener, topicName=topic.name):
        return
    if logger is None:
        logger = logging.getLogger(__name__)
    try:
        pub.subscribe(listener, topic.name)
    except pub.ListenerMismatchError as exc:
        logger.error("ListenerMismatchError")
        logger.error("args: %s", exc.args)
        logger.error("msg: %s", exc.msg)
        logger.error("module: %s", exc.module)
        logger.error("idStr: %s", exc.idStr)
        for topic_listner in pub.getDefaultTopicMgr().getTopic(topic).getListeners():
            logger.error("Topic listener: %s", topic_listner)


def unsubscribe_topic(listener, topic_name):
    """Unsubscribe a listener to a topic and log errors."""
    topic_mgr = pub.getDefaultTopicMgr()
    topic = topic_mgr.getOrCreateTopic(topic_name)
    if pub.isSubscribed(listener, topicName=topic.name):
        pub.unsubscribe(listener, topic.name)


def set_fan_speed(on_level):
    """Map a value to a fan speed."""
    if on_level in FanSpeedRange.OFF.value:
        return FanSpeed.OFF
    if on_level in FanSpeedRange.LOW.value:
        return FanSpeed.LOW
    if on_level in FanSpeedRange.MEDIUM.value:
        return FanSpeed.MEDIUM
    return FanSpeed.HIGH



"""Utilities for managers."""
from ..device_types.ipdb import IPDB
from .device_id_manager import DeviceId


def create_device(device_id: DeviceId):
    """Create an Insteon Device from a DeviceId named Tuple."""

    ipdb = IPDB()
    product = ipdb[[device_id.cat, device_id.subcat]]
    deviceclass = product.deviceclass
    if deviceclass is not None:
        return deviceclass(
            device_id.address,
            device_id.cat,
            device_id.subcat,
            device_id.firmware,
            product.description,
            product.model,
        )
    return None


def create_x10_device(
    housecode: str,
    unitcode: int,
    x10_feature: str,
    steps: int = 22,
    max_level: int = 255,
):
    """Create an Insteon Device from a DeviceId named Tuple."""

    ipdb = IPDB()
    product = ipdb.x10(x10_feature)
    deviceclass = product.deviceclass
    if deviceclass is not None:
        if x10_feature == "dimmable":
            return deviceclass(housecode, unitcode, steps, max_level)
        return deviceclass(housecode, unitcode)
    return None



"""X10 device types."""

from ..aldb.no_aldb import NoALDB
from ..constants import ResponseStatus
from ..x10_address import create


class X10DeviceBase:
    """X10 device base class."""

    def __init__(self, housecode, unitcode):
        """Init the X10DeviceBase class."""

        self._address = create(housecode, unitcode)
        self._aldb = NoALDB(self._address)
        self._description = "Generic X10 device"
        self._model = ""

        self._groups = {}
        self._handlers = {}
        self._managers = {}
        self._events = {}

        self._register_handlers_and_managers()
        self._register_groups()
        self._register_events()
        self._subscribe_to_handelers_and_managers()

    # Public properties
    @property
    def address(self):
        """Return the INSTEON device address."""
        return self._address

    @property
    def description(self):
        """Return the INSTEON device description."""
        return self._description

    @property
    def model(self):
        """Return the INSTEON device model number."""
        return self._model

    @property
    def id(self):
        """Return the ID of the device."""
        return self._address.id

    @property
    def cat(self):
        """Return a fake Insteon device category."""
        return 0xFF

    @property
    def subcat(self):
        """Return a fake Insteon device subcategory."""
        return 0xFF

    @property
    def firmware(self):
        """Return a fake Insteon device firmware."""
        return 0xFF

    @property
    def groups(self):
        """Return the device groups/groups."""
        return self._groups

    @property
    def aldb(self):
        """Return the device All-Link Database."""
        return self._aldb

    @property
    def operating_flags(self):
        """Return the Operating Flags."""
        return {}

    @property
    def properties(self):
        """Return the Extended Properties."""
        return {}

    @property
    def default_links(self):
        """Return the list of default links."""
        return []

    @property
    def is_battery(self):
        """Return True if the device is battery operated."""
        return False

    def status(self, group=None):
        """Get the status of the device."""

    async def async_status(self, group=None):
        """Get the status of the device."""
        return ResponseStatus.SUCCESS

    async def async_read_config(self):
        """Get all configuration settings.

        This command does nothing for X10 devices.

        """
        return ResponseStatus.SUCCESS

    async def async_read_op_flags(self, group=None):
        """Read the device operating flags."""
        return ResponseStatus.SUCCESS

    async def async_write_op_flags(self):
        """Write the operating flags to the device."""
        return ResponseStatus.SUCCESS

    async def async_read_ext_properties(self):
        """Get the device extended properties."""
        return ResponseStatus.SUCCESS

    async def async_write_ext_properties(self):
        """Write the extended properties."""
        return ResponseStatus.SUCCESS

    async def async_read_product_id(self):
        """Get the product ID."""
        return ResponseStatus.SUCCESS

    async def async_add_default_links(self):
        """Add the default links betweent he modem and the device."""
        return ResponseStatus.SUCCESS

    def _register_handlers_and_managers(self):
        """Register handlers and managers."""

    def _register_groups(self):
        """Register groups."""

    def _register_events(self):
        """Register events."""

    def _subscribe_to_handelers_and_managers(self):
        """Subscribe groups and events to handlers and managers."""



"""Base class to handle Broadcast messages from devices."""
from ...address import Address
from ...constants import MessageFlagType
from ..inbound_base import InboundHandlerBase


class AllLinkCleanupCommandHandlerBase(InboundHandlerBase):
    """Base class to handle inbound All-Link Cleanup messages."""

    def __init__(self, topic, address, group):
        """Init the BroadcastHandlerBase class."""
        self._address = Address(address)
        self._group = group
        super().__init__(
            topic=topic,
            address=self._address,
            group=self._group,
            message_type=MessageFlagType.ALL_LINK_CLEANUP,
        )



"""Manage X10 functions."""
from ..utils import subscribe_topic
from ..constants import X10Commands
from ..handlers.to_device.x10_send import X10CommandSend
from ..subscriber_base import SubscriberBase
from ..x10_address import X10Address, create


async def async_x10_all_units_off(housecode: str):
    """Send the X10 All Units OFF command."""
    cmd = X10CommandSend(create(housecode, 0), X10Commands.ALL_UNITS_OFF)
    return await cmd.async_send()


async def async_x10_all_lights_on(housecode: str):
    """Send the X10 All Lights ON command."""
    cmd = X10CommandSend(create(housecode, 0), X10Commands.ALL_LIGHTS_ON)
    return await cmd.async_send()


async def async_x10_all_lights_off(housecode: str):
    """Send the X10 All Lights OFF command."""
    cmd = X10CommandSend(create(housecode, 0), X10Commands.ALL_LIGHTS_OFF)
    return await cmd.async_send()


class X10OnOffManager(SubscriberBase):
    """Manage Inbound X10 On/Off commands."""

    def __init__(self, address: X10Address):
        """Init the X10OnOffManager class."""
        topic = "subscribers.{}.on_off".format(address.id)
        super().__init__(subscriber_topic=topic)
        on_topic = "{}.{}".format(address.id, str(X10Commands.ON))
        off_topic = "{}.{}".format(address.id, str(X10Commands.OFF))
        subscribe_topic(self._on_received, on_topic)
        subscribe_topic(self._off_received, off_topic)

    def subscribe_on(self, callback):
        """Subscribe to ON events."""
        topic = "{}.{}".format(self._subscriber_topic, str(X10Commands.ON))
        subscribe_topic(callback, topic)

    def subscribe_off(self, callback):
        """Subscribe to OFF events."""
        topic = "{}.{}".format(self._subscriber_topic, str(X10Commands.OFF))
        subscribe_topic(callback, topic)

    def _on_received(self):
        """Receive an ON message."""
        self._call_subscribers(on_level=0xFF)

    def _off_received(self):
        """Recieve an OFF message."""
        self._call_subscribers(on_level=0x00)


class X10DimBrightenManager(SubscriberBase):
    """Manage Inbound X10 On/Off commands."""

    def __init__(self, address: X10Address):
        """Init the X10OnOffManager class."""
        topic = "subscribers.{}.dim_bright".format(address.id)
        super().__init__(subscriber_topic=topic)
        dim_topic = "{}.{}".format(address.id, str(X10Commands.DIM))
        bright_topic = "{}.{}".format(address.id, str(X10Commands.BRIGHT))
        subscribe_topic(self._dim_received, dim_topic)
        subscribe_topic(self._bright_received, bright_topic)

    def subscribe_dim(self, callback):
        """Subscribe to DIM events."""
        topic = "{}.{}".format(self._subscriber_topic, str(X10Commands.DIM))
        subscribe_topic(callback, topic)

    def subscribe_bright(self, callback):
        """Subscribe to BRIGHT events."""
        topic = "{}.{}".format(self._subscriber_topic, str(X10Commands.BRIGHT))
        subscribe_topic(callback, topic)

    def _dim_received(self):
        """Receive an ON message."""
        self._call_subscribers(on_level=-1)

    def _bright_received(self):
        """Recieve an OFF message."""
        self._call_subscribers(on_level=1)


class X10AllLightsOnOffManager(SubscriberBase):
    """Manage Inbound X10 On/Off commands."""

    def __init__(self, address: X10Address):
        """Init the X10OnOffManager class."""
        topic = "subscribers.{}.all_lights_on_off".format(address.id)
        super().__init__(subscriber_topic=topic)
        on_topic = "x10{}.{}".format(
            address.housecode.lower(), str(X10Commands.ALL_LIGHTS_ON)
        )
        off_topic = "x10{}.{}".format(
            address.housecode.lower(), str(X10Commands.ALL_LIGHTS_OFF)
        )
        subscribe_topic(self._on_received, on_topic)
        subscribe_topic(self._off_received, off_topic)

    def subscribe_on(self, callback):
        """Subscribe to ON events."""
        topic = "{}.{}".format(self._subscriber_topic, str(X10Commands.ALL_LIGHTS_ON))
        subscribe_topic(callback, topic)

    def subscribe_off(self, callback):
        """Subscribe to OFF events."""
        topic = "{}.{}".format(self._subscriber_topic, str(X10Commands.ALL_LIGHTS_OFF))
        subscribe_topic(callback, topic)

    def _on_received(self):
        """Receive an ON message."""
        self._call_subscribers(on_level=0xFF)

    def _off_received(self):
        """Recieve an OFF message."""
        self._call_subscribers(on_level=0x00)


class X10AllUnitsOffManager(SubscriberBase):
    """Manage Inbound X10 All Units Off commands."""

    def __init__(self, address: X10Address):
        """Init the X10AllUnitsOffManager class."""
        topic = "subscribers.{}.{}".format(address.id, str(X10Commands.ALL_UNITS_OFF))
        super().__init__(subscriber_topic=topic)
        off_topic = "x10{}.{}".format(
            address.housecode.lower(), str(X10Commands.ALL_UNITS_OFF)
        )
        subscribe_topic(self._off_received, off_topic)

    def _off_received(self):
        """Recieve an OFF message."""
        self._call_subscribers(on_level=0x00)



"""Manage inbound ON command from device."""
import logging

from .. import inbound_handler
from ...address import Address
from ...constants import MessageFlagType
from ...topics import THERMOSTAT_SET_POINT_RESPONSE
from ...utils import build_topic
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ThermostatSetPointResponseHandler(InboundHandlerBase):
    """Humidity set point command inbound."""

    def __init__(self, address):
        """Init the ThermostatSetPointResponseHandler class."""
        self._address = Address(address)
        super().__init__(
            topic=THERMOSTAT_SET_POINT_RESPONSE,
            address=self._address,
            message_type=MessageFlagType.DIRECT,
        )
        self._subscriber_topic = build_topic(
            prefix="handler.{}".format(self._address.id),  # Force address
            topic=THERMOSTAT_SET_POINT_RESPONSE,
            message_type=MessageFlagType.DIRECT,
        )

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the Humidity set point response from a device."""
        stage_1_on_minutes = user_data["d3"]
        humidity_high = user_data["d4"]
        humidity_low = user_data["d5"]
        firmwire = user_data["d6"]
        cool_set_point = user_data["d7"]
        heat_set_point = user_data["d8"]
        rf_offset = user_data["d9"]
        self._call_subscribers(
            stage_1_on_minutes=stage_1_on_minutes,
            humidity_high=humidity_high,
            humidity_low=humidity_low,
            firmwire=firmwire,
            cool_set_point=cool_set_point,
            heat_set_point=heat_set_point,
            rf_offset=rf_offset,
        )



"""Modem command to get next ALDB record."""
import logging

from . import ack_handler, nak_handler
from ..topics import GET_NEXT_ALL_LINK_RECORD
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetNextAllLinkRecordHandler(OutboundHandlerBase):
    """Handle Get Next All Link Record commands."""

    def __init__(self):
        """Init the GetNextAldbRecordNak class."""
        super().__init__(topic=GET_NEXT_ALL_LINK_RECORD)

    @ack_handler()
    def handle_ack(self):
        """Handle the ACK message and return True."""

    @nak_handler
    def handle_nak(self):
        """Receive the NAK message and return False."""



"""Handle sending a read request for ALDB records."""
import logging

from .. import ack_handler
from ...address import Address
from ...topics import EXTENDED_GET_SET
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class ExtendedGetCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=EXTENDED_GET_SET, address=address)
        self._data1 = None
        self._data2 = 0x00

    # pylint: disable=arguments-differ
    def send(self, group=0):
        """Send Get Operating Flags message."""
        self._data1 = group
        super().send(data1=self._data1, data2=self._data2)

    # pylint: disable=arguments-differ
    async def async_send(self, group=0):
        """Send Get Operating Flags message asyncronously."""
        self._data1 = group
        response = await super().async_send(data1=self._data1, data2=self._data2)
        self._data1 = None
        return response

    @ack_handler(wait_response=True)
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the ACK response.

        Required to ensure only GET requests are triggered.
        """
        if (
            not user_data
            or not user_data["data1"] == self._data1
            or not user_data["data2"] == self._data2
        ):
            return
        super().handle_ack(cmd1, cmd2, user_data)



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""

from ..default_link import DefaultLink
from ..events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT, Event
from ..groups import DIMMABLE_LIGHT
from ..groups.on_level import OnLevel
from ..handlers.to_device.status_request import StatusRequestCommand
from ..managers.on_level_manager import OnLevelManager
from .commands import STATUS_COMMAND, STATUS_COMMAND_HUB
from .device_base import Device

ON_LEVEL_MANAGER = "on_level_manager"


class VariableControllerBase(Device):
    """Variable state value controller device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the VariableControllerBase class."""
        self._buttons = {1: DIMMABLE_LIGHT} if buttons is None else buttons
        super().__init__(address, cat, subcat, firmware, description, model)

    # pylint: disable=arguments-differ
    async def async_status(self):
        """Request the status of the device.

        Returns a ResponseStatus value
            FAILURE: Device did not acknowledge the message
            SUCCESS: Device acknowledged the message
            UNCLEAR: Device received the message but did not confirm the action

        """
        return await self._handlers[STATUS_COMMAND].async_send()

    def _register_default_links(self):
        """Register default links.

        Default links:
            is_controller group dev_data1 dev_data2 dev_data3 modem_data1 modem_data2 modem_data3

        """
        super()._register_default_links()
        for group in self._buttons:
            link = DefaultLink(
                is_controller=True,
                group=group,
                dev_data1=255,
                dev_data2=28,
                dev_data3=group,
                modem_data1=0,
                modem_data2=0,
                modem_data3=0,
            )
            self._default_links.append(link)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[STATUS_COMMAND] = StatusRequestCommand(self._address)
        # Added for the Hub status request which is often sending cmd2 = 2
        self._handlers[STATUS_COMMAND_HUB] = StatusRequestCommand(
            self._address, status_type=2
        )
        for group in self._buttons:
            if self._managers.get(group) is None:
                self._managers[group] = {}
            self._managers[group][ON_LEVEL_MANAGER] = OnLevelManager(
                self._address, group
            )

    def _register_groups(self):
        super()._register_groups()
        for group in self._buttons:
            name = self._buttons[group]
            self._groups[group] = OnLevel(name=name, address=self._address, group=group)

    def _register_events(self):
        super()._register_events()
        for group in self._buttons:
            if self._events.get(group) is None:
                self._events[group] = {}
            button = self._buttons[group]
            self._events[group][ON_EVENT] = Event(
                ON_EVENT, self._address, group, button
            )
            self._events[group][OFF_EVENT] = Event(
                OFF_EVENT, self._address, group, button
            )
            self._events[group][ON_FAST_EVENT] = Event(
                ON_FAST_EVENT, self._address, group, button
            )
            self._events[group][OFF_FAST_EVENT] = Event(
                OFF_FAST_EVENT, self._address, group, button
            )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[STATUS_COMMAND].subscribe(self._handle_status)
        self._handlers[STATUS_COMMAND_HUB].subscribe(self._handle_status)
        for group in self._buttons:
            state = self._groups[group]
            self._managers[group][ON_LEVEL_MANAGER].subscribe(state.set_value)

            event = self._events[group][ON_EVENT]
            self._managers[group][ON_LEVEL_MANAGER].subscribe_on(event.trigger)

            event = self._events[group][OFF_EVENT]
            self._managers[group][ON_LEVEL_MANAGER].subscribe_off(event.trigger)

            event = self._events[group][ON_FAST_EVENT]
            self._managers[group][ON_LEVEL_MANAGER].subscribe_on_fast(event.trigger)

            event = self._events[group][OFF_FAST_EVENT]
            self._managers[group][ON_LEVEL_MANAGER].subscribe_off_fast(event.trigger)

    def _handle_status(self, db_version, status):
        """Set the status of the dimmable_switch state."""
        self._groups[1].value = status



"""Class to define the message definitions."""


class MessageDefinition:
    """Insteon message defintion."""

    def __init__(self, message_id, message_fields: []):
        """Init the MessageDefinition class."""
        self._message_id = message_id
        self._fields = message_fields

    def __len__(self):
        """Emit the message length."""
        length = 2
        for field in self._fields:
            length += field.length
        return length

    @property
    def message_id(self):
        """Emit the message type."""
        return self._message_id

    @property
    def fields(self):
        """Emit the message fields."""
        return self._fields



"""Manage inbound ON command from device."""
import logging
from collections import OrderedDict

from .. import inbound_handler
from ...address import Address
from ...constants import MessageFlagType
from ...topics import EXTENDED_GET_RESPONSE
from ...utils import build_topic
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ExtendedGetResponseHandler(InboundHandlerBase):
    """Off Level command inbound."""

    def __init__(self, address):
        """Init the OffInbound class."""
        self._address = Address(address)
        super().__init__(
            topic=EXTENDED_GET_RESPONSE,
            address=self._address,
            message_type=MessageFlagType.DIRECT,
        )
        self._subscriber_topic = build_topic(
            prefix="handler.{}".format(self._address),  # Force address
            topic="ext_get_response",
            message_type=MessageFlagType.DIRECT,
        )

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the Extended Get response from a device."""
        data = OrderedDict()
        for item in range(3, 15):
            data["data{}".format(item)] = user_data["d{}".format(item)]
        self._call_subscribers(group=user_data["d1"], data=data)



"""Mock message handling via Chain of Command pattern."""
import asyncio
from abc import ABCMeta

from . import ResponseStatus
from ..utils import publish_topic
from ..utils import build_topic
from .inbound_base import InboundHandlerBase


MSG_TIME = 4  # seconds to send each message in queue, used for timeout below


def _calc_timeout():
    """Calculate the time to wait for a message to be sent."""
    from .. import devices

    if devices and devices.modem:
        return devices.modem.protocol.message_queue.qsize() * MSG_TIME + MSG_TIME
    return MSG_TIME


class OutboundHandlerBase(InboundHandlerBase):
    """Manage a message chain."""

    __meta__ = ABCMeta

    def __init__(self, topic, address=None, group=None, message_type=None):
        """Init the MessageManager."""
        self._message_response = asyncio.Queue()
        self._message_type = message_type
        super().__init__(topic, address=address, group=group, message_type=message_type)

    @property
    def message_response(self) -> asyncio.Queue:
        """Message response queue to manage message status."""
        return self._message_response

    def send(self, **kwargs):
        """Sent the message."""
        asyncio.ensure_future(self.async_send(**kwargs))

    async def async_send(self, **kwargs):
        """Send the message and wait for a status."""

        while not self._message_response.empty():
            # Empty the message queue
            try:
                self._message_response.get_nowait()
            except asyncio.QueueEmpty:
                pass
        send_topic = build_topic(
            topic=self._topic,
            prefix="send",
            address=None,
            group=None,
            message_type=self._message_type,
        )
        publish_topic(send_topic, **kwargs)
        timeout = _calc_timeout()
        try:
            return await asyncio.wait_for(self._message_response.get(), timeout)
        except asyncio.TimeoutError:
            pass
        return ResponseStatus.UNSENT



"""Base object for all command line menues."""

import argparse
import asyncio
import getpass
import inspect
import logging
import os
import sys
from cmd import Cmd
from collections import namedtuple

from .. import devices
from .log_filter import CommandFilter, StdoutFilter
from .utils import patch_stdin_stdout, set_loop, stdio

_LOGGING = logging.getLogger(__name__)
LOG_FILE_NAME = "pyinsteon_tools.log"
STDOUT_LOG_HANDLER = "stdout_handler"
FILE_LOG_HANDLER = "file_handler"
CmdArgs = namedtuple("CmdArgs", "workdir device username host hub_version port")


class ToolsBase(Cmd):
    """Base class for all tools menues."""

    def __init__(self, loop, args=None, menu=None, stdin=None, stdout=None):
        """Init the InsteonCmd class."""
        super().__init__()
        self.stdin = stdin
        self.stdout = stdout
        prompt = "pyinsteon"
        if menu:
            self.prompt = f"{prompt} - {menu}: "
        else:
            self.prompt = f"{prompt}: "
        self._log_prefix = "STDOUT: "
        self.loop = loop
        self.workdir = None

        # connection variables
        self.device = args.device
        self.username = args.username
        self.host = args.host
        self.hub_version = args.hub_version
        def_port = {}
        def_port[1] = 9761
        def_port[2] = 25105
        self.port = args.port if args.port else def_port.get(self.hub_version)

        if hasattr(args, "verbose") and args.verbose:
            self._setup_logging(logging.DEBUG)
        else:
            self._setup_logging(logging.INFO)
        self._add_filter()

        if hasattr(args, "workdir"):
            self.workdir = args.workdir
            if hasattr(args, "logging") and self.workdir:
                self.do_log_to_file("y", self.workdir)

    async def cmdloop(self, intro=None):
        """Override standard cmdloop to make this async.

        Repeatedly issue a prompt, accept input, parse an initial prefix
        off the received input, and dispatch to action methods, passing them
        the remainder of the line as argument.

        """

        self.preloop()
        if not self.stdin and not self.stdout:
            self.stdin, self.stdout = await stdio(loop=self.loop)
            if sys.platform != "win32":
                patch_stdin_stdout(self.stdin, self.stdout)

        try:
            if intro is not None:
                self.intro = intro
            if self.intro:
                self.stdout.write(self.intro + "\n")
            stop = None
            while not stop:
                if self.cmdqueue:
                    line = self.cmdqueue.pop(0)
                else:
                    line = await self._input(prompt=self.prompt)
                    if not line:
                        line = "EOF"
                line = self.precmd(line)
                stop = await self.onecmd(line)
                stop = self.postcmd(stop, line)
            self.postloop()
        finally:
            pass

    async def onecmd(self, line):
        """Override the method to make it async.

        Interpret the argument as though it had been typed in response
        to the prompt.

        This may be overridden, but should not normally need to be;
        see the precmd() and postcmd() methods for useful execution hooks.
        The return value is a flag indicating whether interpretation of
        commands by the interpreter should stop.

        """
        cmd, arg, line = self.parseline(line)
        if not line:
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        self.lastcmd = line

        if line == "EOF":
            self.lastcmd = ""

        if cmd == "":
            return self.default(line)

        try:
            func = getattr(self, "do_" + cmd)
        except AttributeError:
            return self.default(line)
        if inspect.iscoroutinefunction(func) or inspect.isawaitable(func):
            return await func(arg)
        return func(arg)

    # pylint: disable=no-self-use
    def emptyline(self):
        """Change default empty line to do nothing."""
        return

    @classmethod
    def start(cls):
        """Start the loop and the current command set."""
        parser = argparse.ArgumentParser(description=__doc__)
        parser.add_argument("--device", default="", help="Path to PLM device")
        parser.add_argument("--host", default="", help="Hostname or IP address of Hub")
        parser.add_argument("--username", default="", help="Username of Hub")
        parser.add_argument(
            "--hub_version", default="", help="Version of the Hub (1 or 2)"
        )
        parser.add_argument(
            "--port",
            default="",
            help="Hub IP port (Default 25105 for version 2 or 9761 for version 1)",
        )
        parser.add_argument(
            "-v", "--verbose", action="count", help="Set logging level to verbose"
        )
        parser.add_argument("-l", "--logging", action="count", help="Log to file")
        parser.add_argument(
            "--workdir",
            default="",
            help="Working directory for reading and saving device information.",
        )
        args = parser.parse_args()

        set_loop()
        loop = asyncio.get_event_loop()

        intro = "The command line module for pyinsteon is designed to test devices and perform certain common functions."

        try:
            loop.run_until_complete(cls(loop, args).cmdloop(intro=intro))
        except KeyboardInterrupt:
            loop.stop()
            pending = asyncio.Task.all_tasks(loop=loop)
            for task in pending:
                task.cancel()
                try:
                    loop.run_until_complete(task)
                except asyncio.CancelledError:
                    pass
                except KeyboardInterrupt:
                    pass
            loop.close()

    async def start_menu(self):
        """Start the menu."""
        await self.cmdloop()

    # pylint: disable=no-self-use
    def do_list_devices(self, *args, **kwargs):
        """List all devices.

        Usage:
            list_devices
        """
        self._log_command("list_devices")
        self._log_stdout("Address   Cat   Subcat Description")
        self._log_stdout(
            "--------  ----- ------ ------------------------------------------------------------------------"
        )
        for addr in devices:
            device = devices[addr]
            self._log_stdout(
                f"{addr}  0x{device.cat:02x}  0x{device.subcat:02x}   {device.description}"
            )
        self._log_stdout(f"Total devices: {len(devices)}")

    async def do_log_to_file(self, *args, **kwargs):
        """Start logging to file.

        Usage:
            log_to_file y|n
        """
        args = args[0].split()
        try:
            mode = args[0].lower()
        except IndexError:
            mode = None
        if mode not in ["y", "n"]:
            mode = await self._get_char("Log to file (y/n)", values=["y", "n"])

        try:
            if args[1] != "":
                self.workdir = args[1]
        except IndexError:
            pass

        root_logger = logging.getLogger()

        if mode == "n":
            self._log_command("log_to_file n")
            for handler in root_logger.handlers:
                if handler.get_name() == FILE_LOG_HANDLER:
                    root_logger.removeHandler(handler)
            return

        if not self.workdir:
            self.workdir = await self._get_workdir()

        if self.workdir == "":
            self._log_stdout("A value for the working directory is required.")
            return

        log_file = os.path.join(self.workdir, LOG_FILE_NAME)
        self._log_command(f"log_to_file y {self.workdir}")
        file_formatter = logging.Formatter(
            "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
        )
        file_handler = logging.FileHandler(log_file, mode="a")
        file_handler.setFormatter(file_formatter)
        file_handler.set_name(FILE_LOG_HANDLER)
        root_logger.addHandler(file_handler)

    async def do_save_devices(self, *args, **kwargs):
        """Save devices to the working directory.

        Usage:
            save_devices workdir

        workdir: Directory where the saved device file is located
        """
        args = args[0].split()
        try:
            if args[0] != "":
                self.workdir = args[0]
        except IndexError:
            pass

        if not self.workdir:
            self.workdir = await self._get_workdir()

        if self.workdir == "":
            self._log_stdout("A value for the working directory is required.")
            return

        self._log_command(f"save_devices {self.workdir}")
        await devices.async_save(workdir=self.workdir)

    def do_exit(self, *args, **kwargs):
        """Exit the current menu.

        Usage:
            exit
        """
        self._log_command("exit")
        return -1

    async def do_set_log_level(self, *args, **kwargs):
        """Set the log level to INFO (i) or VERBOSE (v).

        Usage:
            set_log_level i|v|
        i: Info
        v: Debug
        m: Show messages
        t: Show topics
        """
        args = args[0].split()
        try:
            mode = args[0].lower()
        except IndexError:
            mode = None

        options = ["i", "v", "m", "t", "n"]
        if mode not in options:
            mode = await self._get_char(
                "Log level (i=info, v=verbose, m=messages, t=topics, n=no messages or topics)",
                values=options,
            )

        self._log_command(f"set_log_level {mode}")
        root_logger = logging.getLogger()
        if mode == "i":
            root_logger.setLevel(logging.INFO)
            message_logger = logging.getLogger("pyinsteon.messages")
            message_logger.setLevel(logging.ERROR)
            topic_logger = logging.getLogger("pyinsteon.topics")
            topic_logger.setLevel(logging.ERROR)
        elif mode == "v":
            root_logger.setLevel(logging.DEBUG)
        elif mode == "m":
            message_logger = logging.getLogger("pyinsteon.messages")
            message_logger.setLevel(logging.DEBUG)
        elif mode == "t":
            topic_logger = logging.getLogger("pyinsteon.topics")
            topic_logger.setLevel(logging.DEBUG)

    async def do_status(self, *args, **kwargs):
        """Display the status of a device.

        Usage:
            status <address>|all y|n

        address: Enter a single address or all for all devices
        reload: y to request the device to send status or
                n to use the current known status
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None
        try:
            refresh_yn = args[1].lower()
            if refresh_yn not in ["y", "n"]:
                refresh_yn = None
        except IndexError:
            refresh_yn = None

        tasks = []
        addresses = await self._get_addresses(
            address=address, allow_all=True, allow_cancel=True
        )
        if not addresses:
            return

        for address in addresses:
            if devices[address] != devices.modem:
                if not refresh_yn:
                    refresh_yn = await self._get_char(
                        "Clear existing records and reload (y/n)",
                        default="n",
                        values=["y", "n"],
                    )
                refresh_current = refresh_yn.lower()
                if refresh_current == "n":
                    for group in devices[address].groups:
                        if devices[address].groups[group].value is None:
                            refresh_current = "y"
                            break
                if refresh_current == "y":
                    tasks.append(devices[address].async_status())
        if tasks:
            await asyncio.gather(*tasks)

        for address in addresses:
            if devices[address] != devices.modem:
                self._print_device_status(address)

    async def do_monitor_mode(self, *args, **kwargs):
        """Enter monitoring mode.

        Usage:
            monitor_mode
        """
        self._log_stdout("Press enter to exit monitor mode")
        self._remove_filter()
        await self.stdin.readline()
        self._add_filter()

    def _print_device_status(self, address):
        """Print device status to log."""
        device = devices[address]
        self._log_stdout("Address  Group State Name      Value")
        self._log_stdout("-------- ----- --------------- -----")
        for group_id in device.groups:
            group = device.groups[group_id]
            self._log_stdout(
                f"{device.address} {group_id:>3d}   {group.name:15.15s} {str(group.value):>5s}"
            )

    def _async_run(self, func, *args, **kwargs):
        """Run a function in the event loop."""
        self.loop.run_until_complete(func(*args, **kwargs))

    async def _get_connection_params(self):
        """Ensure connectoin parameters are filled."""
        password = None
        if not self.device and not self.host:
            self.device = await self._input(
                "USB Device (i.e. /dev/ttyUSB0 or COM5) press enter if Hub: "
            )
        if not self.device:
            if not self.host:
                self.host = await self._input("Hub IP address or hostname: ")
            if not self.username:
                self.username = await self._input("Hub usernme: ")
            password = getpass.getpass(prompt="Hub password: ")
            if not self.hub_version:
                self.hub_version = await self._get_int(
                    "Hub version", default=2, values=[1, 2],
                )
            if not self.port:
                self.port = await self._get_int(
                    "Hub port", default=25105 if self.hub_version == 2 else 9761,
                )
        return password

    async def _input(self, prompt=""):
        """Asyncronous input of a line."""
        self.stdout.write(prompt)
        await self.stdout.drain()
        line = await self.stdin.readline()
        return line.strip("\r\n")

    def _setup_logging(self, level):
        """Set up the initial console logging."""

        root_logger = logging.getLogger()

        # remove any existing logging that are not our handlers
        # add filter to existing stdout handler
        found_our_handlers = False
        for handler in root_logger.handlers:
            if handler.get_name() not in [STDOUT_LOG_HANDLER, FILE_LOG_HANDLER]:
                logging.getLogger().removeHandler(handler)
            else:
                found_our_handlers = True

        if found_our_handlers:
            return

        stdout_formatter = logging.Formatter("%(message)s")
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(stdout_formatter)
        console_handler.set_name(STDOUT_LOG_HANDLER)
        root_logger.addHandler(console_handler)
        root_logger.setLevel(level)
        self._log_stdout(f"Set log level to {level}")

    async def _call_next_menu(self, menu, name=None):
        """Start the next menu."""
        cmd_args = CmdArgs(
            self.workdir,
            self.device,
            self.username,
            self.host,
            self.hub_version,
            self.port,
        )
        await menu(
            self.loop, cmd_args, name, stdin=self.stdin, stdout=self.stdout
        ).start_menu()

    def _add_filter(self):
        """Add a filter for the current menu."""
        root_logger = logging.getLogger()
        found_prompt = False
        found_prefix = False
        for handler in root_logger.handlers:
            if handler.get_name() == STDOUT_LOG_HANDLER:
                for my_filter in handler.filters:
                    if hasattr(my_filter, "prompt") and my_filter.prompt == self.prompt:
                        found_prompt = True
                    if (
                        hasattr(my_filter, "prefix")
                        and my_filter.prefix == self._log_prefix
                    ):
                        found_prefix = True
                if not found_prompt:
                    handler.addFilter(CommandFilter(self.prompt))
                if not found_prefix:
                    handler.addFilter(StdoutFilter(self._log_prefix))

    def _remove_filter(self):
        """Remove the stdout filter to enable log output."""
        root_logger = logging.getLogger()
        found_prefix_filter = False
        for handler in root_logger.handlers:
            if handler.get_name() == STDOUT_LOG_HANDLER:
                for my_filter in handler.filters:
                    if (
                        hasattr(my_filter, "prefix")
                        and my_filter.prefix == self._log_prefix
                    ):
                        found_prefix_filter = True
                        stdout_filter = my_filter
                        break
            if found_prefix_filter:
                handler.removeFilter(stdout_filter)

    def _log_command(self, line):
        """Log the command to the log file if the log file is active."""
        output = f"{self.prompt}{line}"
        _LOGGING.info(output)

    def _log_stdout(self, line):
        """Log a message to standard out."""
        output = f"{self._log_prefix}{line}"
        _LOGGING.info(output)

    async def _get_int(self, prompt, default=None, values=None):
        """Get an integer value."""
        value = None
        if default:
            prompt = f"{prompt} (Default {default}): "
        else:
            prompt = f"{prompt}: "
        while True:
            value = await self._input(prompt)
            if value:
                try:
                    value = int(value)
                    if values and value not in values:
                        raise ValueError()
                    break
                except ValueError:
                    response = "Must be a number."
                    if values:
                        response = f"{response} Acceptable values {values}."
                    self._log_stdout(response)
            elif default is not None:
                value = default
                break
            else:
                response = "A number is required."
                if values:
                    response = f"{response} Acceptable values {values}."
                self._log_stdout(response)
        return value

    async def _get_float(self, prompt, default=None, maximum=None, minimum=None):
        """Get a floating point value."""
        value = None
        if default:
            prompt = f"{prompt} (Default {default}): "
        else:
            prompt = f"{prompt}: "
        while True:
            value = await self._input(prompt)
            if value:
                try:
                    value = float(value)
                    if value < maximum or value > maximum:
                        raise ValueError()
                    break
                except ValueError:
                    response = "Must be a number."
                    if maximum and minimum:
                        response = f"{response} (Max: {maximum}  Min: {minimum})."
                    elif maximum:
                        response = f"{response} (Max: {maximum})."
                    elif maximum and minimum:
                        response = f"{response} (Min: {minimum})."
                    self._log_stdout(response)
            elif default is not None:
                value = default
                break
            else:
                response = "Must be a number."
                if maximum and minimum:
                    response = f"{response} (Max: {maximum}  Min: {minimum})."
                elif maximum:
                    response = f"{response} (Max: {maximum})."
                elif maximum and minimum:
                    response = f"{response} (Min: {minimum})."
                self._log_stdout(response)
        return value

    async def _get_char(self, prompt, default=None, values=None):
        """Get a character string value."""
        if default:
            prompt = f"{prompt} (Default {default.upper()}): "
        else:
            prompt = f"{prompt}: "
        while True:
            value = await self._input(prompt)
            if value:
                if not values or value.lower() in values:
                    break
                self._log_stdout(f"Acceptable values {values}")
            elif default is not None:
                value = default
                break
            else:
                response = "A response is required."
                if values:
                    response = f"{response} Acceptable values {values}."
                self._log_stdout(response)
        return value

    async def _get_workdir(self):
        """Input the valeu for the workdir."""
        self._log_stdout("The working directory stores the lsit of identified devices.")
        self._log_stdout(
            "Enter a working directory where the saved file is (and will be saved to after loading.)"
        )
        workdir = await self._input(
            f"Working directory (enter . for current director): "
        )
        if workdir == ".":
            return os.getcwd()
        return workdir

    async def _get_addresses(
        self,
        address=None,
        allow_cancel=False,
        allow_all=True,
        prompt="Enter device address",
    ):
        """Get the address of a device or all devices."""
        prompt_addr = prompt
        prompt_cancel = f"{prompt} or blank to cancel"
        prompt_all = f"{prompt} or all for all devices"
        prompt_all_cancel = f"{prompt}, all for all devices, or blank to cancel"

        addresses = []
        if allow_all and allow_cancel:
            prompt = f"{prompt_all_cancel}: "
        elif allow_all:
            prompt = f"{prompt_all}: "
        elif allow_cancel:
            prompt = f"{prompt_cancel}: "
        else:
            prompt = prompt_addr

        while True:
            try:
                if not address:
                    address = await self._input(prompt)
                if not address:
                    if allow_cancel:
                        return addresses
                elif str(address).strip("'\"").lower() == "all":
                    for addr in devices:
                        addresses.append(addr)
                    return addresses
                elif devices[address]:
                    addresses.append(address)
                    return addresses
                else:
                    self._log_stdout(f"Device {address} not found in device list.")
                    return []
            except ValueError:
                self._log_stdout("Invalid address entered.")
                address = None

    async def _print_aldb(self, *args):
        """Print the All-Link Database to the log."""
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_all=True, allow_cancel=True
        )
        if not addresses:
            return
        self._log_command(f"print_aldb {'all' if len(addresses) > 1 else addresses[0]}")
        self._print_aldb_out(addresses)

    def _print_aldb_out(self, addresses):
        """Print the ALDB to the log."""
        for address in addresses:
            device = devices[address]
            self._log_stdout("")
            self._log_stdout(
                f"Device: {device.address}  Load Status: {str(device.aldb.status)}"
            )
            self._log_stdout(
                "RecID In Use Mode HWM Group Address  Data 1 Data 2 Data 3"
            )
            self._log_stdout(
                "----- ------ ---- --- ----- -------- ------ ------ ------"
            )
            for mem_addr in device.aldb:
                rec = device.aldb[mem_addr]
                in_use = "Y" if rec.is_in_use else "N"
                mode = "C" if rec.is_controller else "R"
                hwm = "Y" if rec.is_high_water_mark else "N"
                line = f" {rec.mem_addr:04x}    {in_use:s}     {mode:s}   {hwm:s}    {rec.group:3d} {rec.target}   {rec.data1:3d}   {rec.data2:3d}   {rec.data3:3d}"
                self._log_stdout(line)
            self._log_stdout("")



"""Handle sending a read request for ALDB records."""
import logging

from .. import direct_ack_handler
from ...address import Address
from ...topics import GET_OPERATING_FLAGS
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetOperatingFlagsCommand(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=GET_OPERATING_FLAGS, address=address)
        self._group = None

    # pylint: disable=arguments-differ
    def send(self, flags_requested=0):
        """Send Get Operating Flags message."""
        super().send(flags_requested=flags_requested)

    # pylint: disable=arguments-differ
    async def async_send(self, flags_requested=0):
        """Send Get Operating Flags message asyncronously."""
        self._group = flags_requested
        return await super().async_send(flags_requested=self._group)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the direct ACK message."""
        self._call_subscribers(group=self._group, flags=cmd2)
        self._group = None



"""Set the LEDs of a KeyPadLinc."""

from ...constants import MessageFlagType
from ...utils import build_topic
from .extended_set import ExtendedSetCommand


def _bitmask(*args):
    bitmask = 0x00
    for group in range(0, 8):
        val = 1 if args[group] else 0
        bitmask = bitmask + (val << group)
    return bitmask


class SetLedsCommandHandler(ExtendedSetCommand):
    """Set the LEDs of a KeypadLinc."""

    def __init__(self, address):
        """Init the SetLedsCommandHandler class."""
        super().__init__(address=address, data1=0x01, data2=0x09)
        self._subscriber_topic = build_topic(
            prefix="handler.{}".format(self._address.id),  # Force address
            topic="set_leds",
            message_type=MessageFlagType.DIRECT,
        )

    # pylint: disable=arguments-differ
    def send(
        self,
        group1: bool,
        group2: bool,
        group3: bool,
        group4: bool,
        group5: bool,
        group6: bool,
        group7: bool,
        group8: bool,
    ):
        """Set the LED values of the KPL."""
        super().send(
            group1=group1,
            group2=group2,
            group3=group3,
            group4=group4,
            group5=group5,
            group6=group6,
            group7=group7,
            group8=group8,
        )

    # pylint: disable=arguments-differ
    async def async_send(
        self,
        group1: bool,
        group2: bool,
        group3: bool,
        group4: bool,
        group5: bool,
        group6: bool,
        group7: bool,
        group8: bool,
    ):
        """Set the LED values of the KPL."""
        bitmask = _bitmask(
            group1, group2, group3, group4, group5, group6, group7, group8
        )
        return await super().async_send(data3=bitmask)



"""Switched Lighting Control devices (CATEGORY 0x02)."""
from ..events import OFF_EVENT, ON_EVENT
from ..groups import ON_OFF_SWITCH
from ..handlers.to_device.off import OffCommand
from ..handlers.to_device.off_fast import OffFastCommand
from ..handlers.to_device.on_fast import OnFastCommand
from ..handlers.to_device.on_level import OnLevelCommand
from .commands import OFF_COMMAND, OFF_FAST_COMMAND, ON_COMMAND, ON_FAST_COMMAND
from .on_off_controller_base import OnOffControllerBase


class OnOffResponderBase(OnOffControllerBase):
    """Switched Lighting Control device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
        on_event_name=ON_EVENT,
        off_event_name=OFF_EVENT,
        on_fast_event_name=None,
        off_fast_event_name=None,
    ):
        """Init the OnOffResponderBase class."""
        buttons = {1: ON_OFF_SWITCH} if buttons is None else buttons
        super().__init__(
            address,
            cat,
            subcat,
            firmware,
            description,
            model,
            buttons,
            on_event_name,
            off_event_name,
            on_fast_event_name,
            off_fast_event_name,
        )

    def on(self, group: int = 0):
        """Turn on the device."""
        group = 1 if not group else group
        self._handlers[group][ON_COMMAND].send(on_level=0xFF)

    async def async_on(self, group: int = 0):
        """Turn on the device."""
        group = 1 if not group else group
        return await self._handlers[group][ON_COMMAND].async_send(on_level=0xFF)

    def off(self, group: int = 0):
        """Turn off the device."""
        group = 1 if not group else group
        self._handlers[group][OFF_COMMAND].send()

    async def async_off(self, group: int = 0):
        """Turn off the device."""
        group = 1 if not group else group
        return await self._handlers[group][OFF_COMMAND].async_send()

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        for group in self._buttons:
            if self._buttons[group] is not None:
                if self._handlers.get(group) is None:
                    self._handlers[group] = {}
                self._handlers[group][ON_COMMAND] = OnLevelCommand(self._address, group)
                self._handlers[group][OFF_COMMAND] = OffCommand(self._address, group)
                self._handlers[group][ON_FAST_COMMAND] = OnFastCommand(
                    self._address, group
                )
                self._handlers[group][OFF_FAST_COMMAND] = OffFastCommand(
                    self._address, group
                )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        for group in self._buttons:
            if self._groups.get(group):
                self._handlers[group][ON_COMMAND].subscribe(
                    self._groups[group].set_value
                )
                self._handlers[group][OFF_COMMAND].subscribe(
                    self._groups[group].set_value
                )
                self._handlers[group][ON_FAST_COMMAND].subscribe(
                    self._groups[group].set_value
                )
                self._handlers[group][OFF_FAST_COMMAND].subscribe(
                    self._groups[group].set_value
                )

            if self._on_event_name:
                event = self._events[group][self._on_event_name]
                self._handlers[group][ON_COMMAND].subscribe(event.trigger)

            if self._off_event_name:
                event = self._events[group][self._off_event_name]
                self._handlers[group][OFF_COMMAND].subscribe(event.trigger)

            if self._on_fast_event_name:
                event = self._events[group][self._on_fast_event_name]
                self._handlers[group][ON_FAST_COMMAND].subscribe(event.trigger)

            if self._off_fast_event_name:
                event = self._events[group][self._off_fast_event_name]
                self._handlers[group][OFF_FAST_COMMAND].subscribe(event.trigger)



"""Temperature state."""
from .group_base import GroupBase
from ..address import Address
from ..constants import ThermostatMode


class Temperature(GroupBase):
    """Temperature state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: float = None,
    ):
        """Init the Temperature class."""
        super().__init__(name, address, group, default, value_type=float)

    # pylint: disable=arguments-differ
    def set_value(self, temperature):
        """Set the temperature value."""
        self.value = temperature

    @property
    def value(self):
        """Return the temperature value."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the temperature value."""
        try:
            self._value = self._type(value) if value is not None else None
        except TypeError:
            raise TypeError(
                "Error setting value of State {}: "
                "Must be of type {}".format(self._name, self._type.__name__)
            )
        else:
            self._call_subscribers(
                name=self._name,
                address=self._address.id,
                value=self._value,
                group=self._group,
            )


class Humidity(GroupBase):
    """Humidity state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: float = None
    ):
        """Init the Temperature class."""
        super().__init__(name, address, group, default, value_type=float)

    # pylint: disable=arguments-differ
    def set_value(self, humidity):
        """Set the temperature value."""
        self.value = humidity


class SetPoint(GroupBase):
    """SetPont state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: float = None
    ):
        """Init the SetPoint class."""
        super().__init__(name, address, group, default, value_type=float)

    # pylint: disable=arguments-differ
    def set_value(self, degrees):
        """Set the Set Point value."""
        self.value = degrees


class SystemMode(GroupBase):
    """System Mode state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: ThermostatMode = None,
    ):
        """Init the SystemMode class."""
        super().__init__(name, address, group, default, value_type=ThermostatMode)

    # pylint: disable=arguments-differ
    def set_value(self, mode: ThermostatMode):
        """Set the system mode value."""
        self.value = mode

    @property
    def value(self):
        """Return the group value."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            if value > 0x0F:
                value = value >> 4
            self._value = self._type(value) if value is not None else None
        except TypeError:
            raise TypeError(
                "Error setting value of State {}: "
                "Must be of type {}".format(self._name, self._type.__name__)
            )
        else:
            self._call_subscribers(
                name=self._name,
                address=self._address.id,
                value=self._value,
                group=self._group,
            )


class FanMode(GroupBase):
    """Fan Mode state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: ThermostatMode = None,
    ):
        """Init the SystemMode class."""
        super().__init__(name, address, group, default, value_type=ThermostatMode)

    # pylint: disable=arguments-differ
    def set_value(self, mode: ThermostatMode):
        """Set the system mode value."""
        self.value = mode

    @property
    def value(self):
        """Return the group value."""
        return self._value

    @value.setter
    def value(self, value):
        """Set the value of the state."""
        try:
            if value > 0x0F:
                value = value & 0x0F
            self._value = self._type(value) if value is not None else None
        except TypeError:
            raise TypeError(
                "Error setting value of State {}: "
                "Must be of type {}".format(self._name, self._type.__name__)
            )
        else:
            self._call_subscribers(
                name=self._name,
                address=self._address.id,
                value=self._value,
                group=self._group,
            )



"""Insteon message and command handlers."""
import asyncio
import logging
from datetime import datetime
from functools import wraps

from ..utils import subscribe_topic
from ..address import Address
from ..constants import MessageFlagType, ResponseStatus
from ..utils import build_topic

TIMEOUT = 3  # Time out between ACK and Direct ACK
_LOGGER = logging.getLogger(__name__)


async def _async_post_response(
    obj, response: ResponseStatus, func=None, args=None, kwargs=None
):
    """Post a response status to the resonse queue."""
    if hasattr(obj, "response_lock"):
        if obj.response_lock.locked():
            await obj.message_response.put(response)
            if func is not None:
                func(obj, *args, **kwargs)
        if obj.response_lock.locked():
            obj.response_lock.release()
    else:
        await obj.message_response.put(response)
        if func is not None:
            func(obj, *args, **kwargs)


def _register_handler(
    instance_func, topic, prefix=None, address=None, group=None, message_type=None
):
    full_topic = build_topic(
        topic=topic,
        prefix=prefix,
        address=address,
        group=group,
        message_type=message_type,
    )
    subscribe_topic(instance_func, full_topic)


def inbound_handler(func):
    """Register any inbound message handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=None,
            message_type=message_type,
        )

    func.register_topic = register_topic
    return func


def ack_handler(wait_response=False, timeout=TIMEOUT):
    """Register the message ACK handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound ACK message."""
        _register_handler(
            instance_func,
            topic,
            prefix="ack",
            address=address,
            group=group,
            message_type=message_type,
        )

    async def _wait_response(lock: asyncio.Lock, queue: asyncio.Queue):
        """Wait for the direct ACK message, and post False if timeout reached."""
        # TODO: Need to consider the risk of this. We may be unlocking a prior send command.
        # This would mean that the prior command will terminate. What happens when the
        # prior command returns a direct ACK then this command returns a direct ACK?
        # Do not believe this is an issue but need to test.
        if lock.locked():
            lock.release()
        await lock.acquire()
        try:
            await asyncio.wait_for(lock.acquire(), TIMEOUT)
        except asyncio.TimeoutError:
            if lock.locked():
                await queue.put(ResponseStatus.FAILURE)
        if lock.locked():
            lock.release()

    def setup(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if hasattr(self, "group"):
                group = (
                    1 if not kwargs.get("user_data") else kwargs.get("user_data")["d1"]
                )
                if self.group != group:
                    return
            if wait_response:
                asyncio.ensure_future(
                    _wait_response(self.response_lock, self.message_response)
                )
            else:
                asyncio.ensure_future(
                    _async_post_response(self, ResponseStatus.SUCCESS)
                )
            return func(self, *args, **kwargs)

        wrapper.register_topic = register_topic
        return wrapper

    return setup


def nak_handler(func):
    """Register the message NAK handler.

    This should only be used if the command requires a special NAK
    handler. Under normal conditions all NAK responses are resent by the
    Protocol.
    """

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound ACK message."""
        _register_handler(
            instance_func,
            topic,
            prefix="nak",
            address=address,
            group=group,
            message_type=message_type,
        )

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        asyncio.ensure_future(
            _async_post_response(self, ResponseStatus.FAILURE, func, args, kwargs)
        )

    wrapper.register_topic = register_topic
    return wrapper


def direct_ack_handler(func):
    """Register the DIRECT_ACK response handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound ACK message."""
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=None,
            message_type=MessageFlagType.DIRECT_ACK,
        )

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        asyncio.ensure_future(
            _async_post_response(self, ResponseStatus.SUCCESS, func, args, kwargs)
        )

    wrapper.register_topic = register_topic
    return wrapper


def status_ack_handler(timeout=TIMEOUT):
    """Register the message ACK handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound ACK message."""
        _register_handler(
            instance_func,
            topic,
            prefix="ack",
            address=address,
            group=group,
            message_type=message_type,
        )

    async def _wait_response(self):
        """Wait for the direct ACK message, and post False if timeout reached."""
        # TODO: Need to consider the risk of this. We may be unlocking a prior send command.
        # This would mean that the prior command will terminate. What happens when the
        # prior command returns a direct ACK then this command returns a direct ACK?
        # Do not believe this is an issue but need to test.
        if self.response_lock.locked():
            self.response_lock.release()
        await self.response_lock.acquire()
        try:
            await asyncio.wait_for(self.response_lock.acquire(), TIMEOUT)
        except asyncio.TimeoutError:
            if self.response_lock.locked():
                await self.message_response.put(ResponseStatus.FAILURE)
        if self.response_lock.locked():
            self.response_lock.release()
        self.status_active = False

    def setup(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            cmd2 = kwargs.get("cmd2")
            if cmd2 == self.status_type:
                self.status_active = True
                asyncio.ensure_future(_wait_response(self))
                return func(self, *args, **kwargs)

        wrapper.register_topic = register_topic
        return wrapper

    return setup


def status_handler(func):
    """Register the status response handler."""

    def register_status(instance_func, address):
        # This registers all messages for a device but only triggers on
        # status messages if they return within the TIMEOUT period
        address = Address(address)
        subscribe_topic(instance_func, address.id)

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.status_active:
            asyncio.ensure_future(
                _async_post_response(self, ResponseStatus.SUCCESS, func, args, kwargs)
            )

    wrapper.register_status = register_status
    return wrapper


def direct_nak_handler(func):
    """Register the DIRECT_NAK response handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound DIRECT NAK message."""
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=None,
            message_type=MessageFlagType.DIRECT_NAK,
        )

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        asyncio.ensure_future(
            _async_post_response(self, ResponseStatus.UNCLEAR, func, args, kwargs)
        )

    wrapper.register_topic = register_topic
    return wrapper


def broadcast_handler(func):
    """Register the BROADCAST message handler."""
    last_command = datetime(1, 1, 1)
    last_hops_left = None

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound BROADCAST message."""
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=group,
            message_type=MessageFlagType.BROADCAST,
        )
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        nonlocal last_command, last_hops_left
        curr_time = datetime.now()
        tdelta = curr_time - last_command
        last_command = curr_time
        hops_left = kwargs["hops_left"]
        if last_hops_left is None or hops_left >= last_hops_left or tdelta.seconds >= 2:
            last_hops_left = hops_left
            return func(self, *args, **kwargs)
        last_hops_left = hops_left

    wrapper.register_topic = register_topic
    return wrapper


def all_link_cleanup_ack_handler(func):
    """Register the all_link_cleanup ACK response handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound BROADCAST message."""
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=None,
            message_type=MessageFlagType.ALL_LINK_CLEANUP_ACK,
        )

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        asyncio.ensure_future(
            _async_post_response(self, ResponseStatus.FAILURE, func, args, kwargs)
        )

    wrapper.register_topic = register_topic
    return wrapper


def all_link_cleanup_nak_handler(func):
    """Register the all_link_cleanup NAK response handler."""

    def register_topic(
        instance_func, topic, address=None, group=None, message_type=None
    ):
        """Register an inbound BROADCAST message."""
        _register_handler(
            instance_func,
            topic,
            prefix=None,
            address=address,
            group=None,
            message_type=MessageFlagType.ALL_LINK_CLEANUP_NAK,
        )

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        asyncio.ensure_future(
            _async_post_response(self, ResponseStatus.FAILURE, func, args, kwargs)
        )

    wrapper.register_topic = register_topic
    return wrapper



"""Insteon device types."""

from .unknown_device import UnknownDevice
from .hub import Hub
from .plm import PLM
from .general_controller import (
    GeneralController,
    GeneralController_ControlLinc,
    GeneralController_RemoteLinc,
    GeneralController_MiniRemote_Switch,
    GeneralController_MiniRemote_4,
    GeneralController_MiniRemote_8,
)
from .dimmable_lighting_control import (
    DimmableLightingControl,
    DimmableLightingControl_LampLinc,
    DimmableLightingControl_SwitchLinc,
    DimmableLightingControl_InLineLinc,
    DimmableLightingControl_OutletLinc,
    DimmableLightingControl_DinRail,
    DimmableLightingControl_FanLinc,
    DimmableLightingControl_KeypadLinc_6,
    DimmableLightingControl_KeypadLinc_8,
    DimmableLightingControl_ToggleLinc,
)
from .switched_lighting_control import (
    SwitchedLightingControl,
    SwitchedLightingControl_ApplianceLinc,
    SwitchedLightingControl_SwitchLinc,
    SwitchedLightingControl_InLineLinc,
    SwitchedLightingControl_OutletLinc,
    SwitchedLightingControl_DinRail,
    SwitchedLightingControl_KeypadLinc_6,
    SwitchedLightingControl_KeypadLinc_8,
    SwitchedLightingControl_OnOffOutlet,
    SwitchedLightingControl_ToggleLinc,
)

from .climate_control import (
    ClimateControl_Thermostat,
    ClimateControl_WirelessThermostat,
)

from .security_health_safety import (
    SecurityHealthSafety,
    SecurityHealthSafety_OpenCloseSensor,
    SecurityHealthSafety_MotionSensor,
    SecurityHealthSafety_DoorSensor,
    SecurityHealthSafety_LeakSensor,
    SecurityHealthSafety_Smokebridge,
)

from .sensors_actuators import SensorsActuators, SensorsActuators_IOLink

from .window_coverings import WindowCovering

from .x10 import X10OnOff, X10Dimmable, X10OnOffSensor



"""Device manager."""
import logging
import asyncio

from ..address import Address
from ..x10_address import X10Address
from ..device_types.device_base import Device
from ..device_types.x10_base import X10DeviceBase
from ..device_types.modem_base import ModemBase
from ..managers.saved_devices_manager import SavedDeviceManager
from ..subscriber_base import SubscriberBase
from .device_id_manager import DeviceId, DeviceIdManager
from .device_link_manager import DeviceLinkManager
from .utils import create_device, create_x10_device

DEVICE_INFO_FILE = "insteon_devices.json"
_LOGGER = logging.getLogger(__name__)


# TODO remove devices
class DeviceManager(SubscriberBase):
    """Manages the list of active devices."""

    def __init__(self):
        """Init the DeviceManager class."""
        super().__init__(subscriber_topic="device_added")
        self._devices = {}
        self._modem = None
        self._id_manager = DeviceIdManager()
        self._id_manager.subscribe(self._device_identified)
        self._loading_saved_lock = asyncio.Lock()
        self._link_manager = DeviceLinkManager(self)

    def __getitem__(self, address) -> Device:
        """Return a a device from the device address."""
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        return self._devices.get(address)

    def __iter__(self):
        """Return an iterator of device addresses."""
        for address in self._devices:
            yield address

    def __setitem__(self, address, device):
        """Add a device to the device list."""
        _LOGGER.info("Adding device to INSTEON devices list: %s", address.id)
        if not isinstance(device, (Device, DeviceId, X10DeviceBase)):
            raise ValueError("Device must be a DeviceId or a Device type.")

        if isinstance(device, DeviceId):
            device = create_device(device)

        self._devices[device.address] = device
        if isinstance(device, Device):
            self._id_manager.set_device_id(
                device.address, device.cat, device.subcat, device.firmware
            )
        self._call_subscribers(address=device.address.id)

    def __len__(self):
        """Return the number of devices."""
        return len(self._devices)

    def get(self, address) -> Device:
        """Return a device from an address."""
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        return self._devices.get(address)

    @property
    def modem(self):
        """Return the Insteon Modem."""
        return self._modem

    @modem.setter
    def modem(self, modem):
        """Set the Insteon Modem."""
        if not isinstance(modem, ModemBase):
            raise ValueError("Must be an Insteon Modem object")
        self._modem = modem
        self._devices[self._modem.address] = self._modem

    @property
    def id_manager(self):
        """Return the ID manager instance."""
        return self._id_manager

    def set_id(self, address: Address, cat: int, subcat: int, firmware: int):
        """Add a device override to identify the device information.

        Typical use is to identify devices that do not respond to an ID Request
        such as a battery opperated device.

        """
        address = Address(address)
        device = self[address]
        if device and device.cat == cat and device.subcat == subcat:
            return
        self._id_manager.set_device_id(address, cat, subcat, firmware)

    def add_x10_device(
        self,
        housecode: str,
        unitcode: int,
        x10_feature: str,
        steps: int = 22,
        max_level: int = 255,
    ):
        """Add an X10 device."""
        device = create_x10_device(housecode, unitcode, x10_feature, steps, max_level)
        if device:
            self[device.address] = device
        return device

    async def async_close(self):
        """Close the device ID listener."""
        self._id_manager.close()

    async def async_load(self, workdir="", id_devices=1, load_modem_aldb=1):
        """Load devices from the `insteon_devices.yaml` file and device overrides.

        Parameters:
            workdir: Directory name to find the `insteon_devices.json` file

            id_devices: Indicate if devices should be identified using ID Request
                0: No devices are identified
                1: Unknown devices are identified
                2: All devices are identified
            (default=1)

            load_modem_aldb: Indicate if the Modem ALDB should be loaded
                0: Do not load
                1: Load if not loaded from save file
                2: Load
            (default=1)

        The Modem ALDB is loaded if `refresh` is True or if the saved file has no devices.

        """
        if workdir:
            await self._loading_saved_lock.acquire()
            saved_devices_manager = SavedDeviceManager(workdir, self.modem)
            devices = await saved_devices_manager.async_load()
            for address in devices:
                self[address] = devices[address]
            if self._loading_saved_lock.locked():
                self._loading_saved_lock.release()

        if load_modem_aldb == 0:
            load_modem_aldb = False
        elif load_modem_aldb == 2:
            load_modem_aldb = True
        else:
            load_modem_aldb = not self._modem.aldb.is_loaded

        if load_modem_aldb:
            await self._modem.aldb.async_load()

        for mem_addr in self._modem.aldb:
            rec = self._modem.aldb[mem_addr]
            if rec.target != Address("000000"):
                self._id_manager.append(rec.target)

        if id_devices:
            id_all = id_devices == 2
            await self._id_manager.async_id_devices(refresh=id_all)

    async def async_save(self, workdir):
        """Save devices to a device information file."""
        saved_devices_manager = SavedDeviceManager(workdir, self.modem)
        await saved_devices_manager.async_save(self._devices)

    def _device_identified(self, device_id: DeviceId):
        """Device identified by device ID manager."""
        if self._loading_saved_lock.locked():
            return
        if device_id.cat is not None:
            device = create_device(device_id)
            if self[device_id.address]:
                # If the device is already in the devices list and the cat and subcat
                # are the same, do not add the device again
                if (
                    device_id.cat == self[device_id.address].cat
                    and device_id.subcat == self._devices[device_id.address].subcat
                ):
                    return
            self[device_id.address] = device
            if device_id.cat != 0x03:
                asyncio.ensure_future(device.async_get_engine_version())
                asyncio.ensure_future(self.async_setup_device(device))
            _LOGGER.debug("Device %s added", device.address)

    async def async_setup_device(self, device):
        """Set up device."""
        await device.aldb.async_load(refresh=True)
        await device.async_read_op_flags()
        await device.async_read_ext_properties()
        await device.async_add_default_links()



"""Insteon All-Link Database.

The All-Link database contains database records that represent links to other
Insteon devices that either respond to or control the current device.
"""
import asyncio
import logging
from typing import Callable

from ..address import Address
from ..constants import ALDBStatus, ALDBVersion
from .aldb_record import ALDBRecord
from .aldb_base import ALDBBase
from ..managers.aldb_read_manager import ALDBReadManager
from ..managers.aldb_write_manager import ALDBWriteManager
from ..handlers import ResponseStatus

_LOGGER = logging.getLogger(__name__)


class ALDB(ALDBBase):
    """All-Link Database for a device."""

    def __init__(self, address, version=ALDBVersion.V2, mem_addr=0x0FFF):
        """Init the ALDB class."""

        super().__init__(address=address, version=version, mem_addr=mem_addr)
        self._read_manager = ALDBReadManager(self)
        self._write_manager = ALDBWriteManager(self)

    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""
        if not isinstance(record, ALDBRecord):
            raise ValueError

        self._dirty_records.append(record)

    def load(self, refresh=False, callback: Callable = None):
        """Load the ALDB calling the callback when done."""
        self._cb_aldb_loaded = callback
        asyncio.ensure_future(self.async_load(refresh))

    # pylint: disable=arguments-differ
    async def async_load(
        self,
        mem_addr: int = 0x00,
        num_recs: int = 0x00,
        refresh: bool = False,
        callback: Callable = None,
    ):
        """Load the All-Link Database."""
        _LOGGER.debug("Loading the ALDB async")
        self._status = ALDBStatus.LOADING
        if refresh:
            self.clear()
        # pylint: disable=not-an-iterable
        async for rec in self._read_manager.async_read(
            mem_addr=mem_addr, num_recs=num_recs
        ):
            if self._hwm_record and rec.mem_addr < self._hwm_record.mem_addr:
                continue
            if rec.mem_addr > self.first_mem_addr:
                continue
            if self._records.get(rec.mem_addr):
                self._notify_change(self._records[rec.mem_addr], force_delete=True)
            self._records[rec.mem_addr] = rec
            if rec.is_high_water_mark:
                self._hwm_record = rec
            self._notify_change(rec)
        self.set_load_status()
        if callback:
            callback()
        return self._status

    def write(self, force=False):
        """Write modified records to the device."""
        asyncio.ensure_future(self.async_write(force=force))

    async def async_write(self, force=False):
        """Write modified records to the device.

        Returns a tuple of (completed, failed) record counts.
        """

        if not self.is_loaded and not force:
            _LOGGER.warning(
                "ALDB must be loaded before it can be written Status: %s",
                str(self._status),
            )
            return

        completed = []
        failed = []
        for record in self._dirty_records:
            if record.mem_addr == 0x0000:
                mem_addr = self._existing_link(
                    record.is_controller, record.group, record.target
                )
                if mem_addr is None:
                    _LOGGER.debug("Existing record not found")
                    mem_addr = self._get_next_mem_addr()
                    _LOGGER.debug("Using new record %04x", mem_addr)
                else:
                    _LOGGER.debug("Using existing record %04x", mem_addr)
                record.mem_addr = mem_addr
            # We assume a direct ACK is a confirmation of write.
            # Should we re-read to ensure it is correct.
            response = ResponseStatus.UNSENT
            retries = 0
            while response != ResponseStatus.SUCCESS and retries < 3:
                response = await self._write_manager.async_write(record)
                _LOGGER.debug("Response: %s", str(response))
                retries += 1
            if response == ResponseStatus.SUCCESS:
                self._records[record.mem_addr] = record
                completed.append(record)
            else:
                failed.append(record)
        self._dirty_records = failed
        return len(completed), len(self._dirty_records)

    def add(
        self,
        group: int,
        target: Address,
        controller: bool = False,
        data1: int = 0x00,
        data2: int = 0x00,
        data3: int = 0x00,
        bit5: int = True,
        bit4: int = False,
    ):
        """Add an All-Link record.

        This method does not write to the device. To write modifications to the device
        use `write` or `async_write`.
        """
        mem_addr = 0x0000

        rec = ALDBRecord(
            memory=mem_addr,
            in_use=True,
            controller=controller,
            high_water_mark=False,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            bit5=bit5,
            bit4=bit4,
        )
        self._dirty_records.append(rec)

    def remove(self, mem_addr: int):
        """Remove an All-Link record."""
        rec = self._records.get(mem_addr)
        if not rec:
            raise IndexError("Memory location not found.")
        new_rec = ALDBRecord(
            memory=rec.mem_addr,
            in_use=False,
            controller=rec.is_controller,
            high_water_mark=rec.is_high_water_mark,
            group=rec.group,
            target=rec.target,
            data1=rec.data1,
            data2=rec.data2,
            data3=rec.data3,
            bit5=rec.is_bit5_set,
            bit4=rec.is_bit4_set,
        )
        self._dirty_records.append(new_rec)

    def modify(
        self,
        mem_addr: int,
        in_use: bool = None,
        group: int = None,
        controller: bool = None,
        high_water_mark: bool = None,
        target: Address = None,
        data1: int = None,
        data2: int = None,
        data3: int = None,
        bit5: bool = None,
        bit4: bool = None,
    ):
        """Modify an All-Link record."""
        rec = self._records.get(mem_addr)
        if not rec:
            raise IndexError("Memory location not found.")

        new_in_use = rec.in_use if in_use is None else bool(in_use)
        new_group = rec.group if group is None else int(group)
        new_controller = rec.is_controller if controller is None else bool(controller)
        new_high_water_mark = (
            rec.is_high_water_mark if high_water_mark is None else bool(high_water_mark)
        )
        new_target = rec.target if target is None else Address(target)
        new_data1 = rec.data1 if data1 is None else data1
        new_data2 = rec.data2 if data2 is None else data2
        new_data3 = rec.data3 if data3 is None else data3
        new_bit5_set = rec.is_bit5_set if bit5 is None else bool(bit5)
        new_bit4_set = rec.is_bit4_set if bit4 is None else bool(bit4)
        new_rec = ALDBRecord(
            memory=rec.mem_addr,
            in_use=new_in_use,
            controller=new_controller,
            high_water_mark=new_high_water_mark,
            group=new_group,
            target=new_target,
            data1=new_data1,
            data2=new_data2,
            data3=new_data3,
            bit5=new_bit5_set,
            bit4=new_bit4_set,
        )
        self._dirty_records.append(new_rec)

    def _existing_link(self, is_controller, group, address):
        """Test if a link exists in a device ALDB."""
        for mem_addr in self._records:
            rec = self._records[mem_addr]
            if (
                rec.is_controller == is_controller
                and rec.target == address
                and rec.group == group
            ):
                return rec.mem_addr
        return None

    def _get_next_mem_addr(self):
        """Return the next memory slot available."""
        if not self.is_loaded:
            return None

        last_record = None
        for mem_addr in self._records:
            record = self._records[mem_addr]
            last_record = record
            if not record.is_in_use:
                _LOGGER.debug("Using the available record %04x", record.mem_addr)
                return record.mem_addr

        _LOGGER.debug("Using the next vailable record %04x", record.mem_addr)
        return last_record.mem_addr - 8

    def _calc_load_status(self):
        """Test if the ALDB is fully loaded."""
        has_first = False
        has_last = False
        has_all = False
        last_addr = 0x0000
        first = True
        for mem_addr in sorted(self._records, reverse=True):
            if first:
                _LOGGER.debug("First Addr: 0x%4x", mem_addr)
            if mem_addr == self._mem_addr:
                has_first = True
            if self._records[mem_addr].is_high_water_mark:
                has_last = True
            if last_addr != 0x0000:
                has_all = (last_addr - mem_addr) == 8
            last_addr = mem_addr
        _LOGGER.debug("Has First is %s", has_first)
        _LOGGER.debug("Has Last is %s", has_last)
        _LOGGER.debug("Has All is %s", has_all)
        return has_first and has_all and has_last

    def set_load_status(self):
        """Review the ALDB records and identify the load status."""
        _LOGGER.debug("Setting the load status")
        if self._calc_load_status():
            self._status = ALDBStatus.LOADED
        elif self._records:
            self._status = ALDBStatus.PARTIAL
        else:
            self._status = ALDBStatus.EMPTY

    def _confirm_hwm(self, rec):
        """Confirm the new record is not below the High Wter Mark.

        The ALDB will often respond with records that are below the HWM.
        This method confirms no records below the HWM are added.
        If a new HWM is received, this will also remove records below it.

        Records below the HWM are never used by the device and are therefore
        irrelivant.
        """
        curr_hwm_mem_addr = 0x0000
        for curr_mem_addr in self._records:
            curr_rec = self._records[curr_mem_addr]
            if curr_rec.is_high_water_mark:
                curr_hwm_mem_addr = curr_mem_addr
                break

        found_hwm = (
            curr_hwm_mem_addr != 0x0000
            and self._records[curr_hwm_mem_addr].is_high_water_mark
        )

        if rec.is_high_water_mark and rec.mem_addr < curr_hwm_mem_addr:
            curr_hwm_mem_addr = rec.mem_addr

        elif found_hwm and rec.is_high_water_mark and rec.mem_addr > curr_hwm_mem_addr:
            return False

        remove_records = []
        for curr_mem_addr in self._records:
            if found_hwm and curr_mem_addr < curr_hwm_mem_addr:
                remove_records.append(curr_mem_addr)

        for mem_addr in remove_records:
            self._records.pop(mem_addr)

        return True



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...topics import EXTENDED_GET_SET_2
from .direct_command import DirectCommandHandlerBase
from .. import direct_ack_handler

_LOGGER = logging.getLogger(__name__)


class ExtendedGet2Command(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=EXTENDED_GET_SET_2, address=address)
        self._data1 = 0x00

    # pylint: disable=arguments-differ
    def send(self, group=0):
        """Send Get Operating Flags message."""
        self._data1 = group
        super().send(data1=self._data1)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send Get Operating Flags message asyncronously."""
        return await super().async_send(data1=self._data1)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the direct ACK."""
        self._call_subscribers()



"""Assign a device to an all link group."""
from .. import broadcast_handler
from ...address import Address
from ...topics import DELETE_FROM_ALL_LINK_GROUP
from .broadcast_command import BroadcastCommandHandlerBase


class DeleteFromAllLinkGroupCommand(BroadcastCommandHandlerBase):
    """Delete From All-Link Group command handler."""

    def __init__(self, address: Address):
        """Init the DeleteFromAllLinkGroupCommand class."""
        self._address = Address(address)
        super().__init__(topic=DELETE_FROM_ALL_LINK_GROUP, address=self._address)

    @broadcast_handler
    def receive_id(self, cmd1, cmd2, target, user_data, hops_left):
        """Receive the device ID information."""
        cat = target.high
        subcat = target.middle
        firmware = target.low
        self._call_subscribers(
            address=self._address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            group=cmd2,
            mode=None,
        )



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""
from ..default_link import DefaultLink
from ..events import CLOSE_EVENT, OPEN_EVENT, Event
from ..groups import OPEN_CLOSE_SENSOR
from ..groups.open_close import NormallyClosed, NormallyOpen
from ..handlers.to_device.status_request import StatusRequestCommand
from ..managers.on_level_manager import OnLevelManager
from .commands import STATUS_COMMAND
from .device_base import Device


class OpenCloseControllerBase(Device):
    """Base device for Open/Close controllers."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        normally_open=True,
        state_name=OPEN_CLOSE_SENSOR,
        open_event_name=OPEN_EVENT,
        close_event_name=CLOSE_EVENT,
    ):
        """Init the OpenCloseControllerBase class."""
        self._state_name = state_name
        self._open_event_name = open_event_name
        self._close_event_name = close_event_name
        self._normally_open = normally_open
        super().__init__(address, cat, subcat, firmware, description, model)

    def status(self, group=None):
        """Request the status of the device."""
        self._handlers[STATUS_COMMAND].send()

    async def async_status(self, group=None):
        """Request the status of the device."""
        return await self._handlers[STATUS_COMMAND].async_send()

    def _register_default_links(self):
        super()._register_default_links()
        link = DefaultLink(
            is_controller=True,
            group=1,
            dev_data1=255,
            dev_data2=28,
            dev_data3=1,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        self._default_links.append(link)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[STATUS_COMMAND] = StatusRequestCommand(self._address, 0)
        self._managers[1] = OnLevelManager(self._address, 1)

    def _register_groups(self):
        """Register a Normally Open state."""
        if self._normally_open:
            self._groups[1] = NormallyOpen(self._state_name, self._address, 1)
        else:
            self._groups[1] = NormallyClosed(self._state_name, self._address, 1)

    def _register_events(self):
        self._events[1] = {}
        self._events[1][self._open_event_name] = Event(
            self._open_event_name, self._address, 1
        )
        self._events[1][self._close_event_name] = Event(
            self._close_event_name, self._address, 1
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[STATUS_COMMAND].subscribe(self._handle_status)
        self._managers[1].subscribe(self._groups[1].set_value)
        if self._normally_open:
            # Open is OFF and Close is ON
            self._managers[1].subscribe_off(self._events[self._open_event_name].trigger)
            self._managers[1].subscribe_on(self._events[self._close_event_name].trigger)
        else:
            # Close is OFF and Open is ON
            self._managers[1].subscribe_on(self._events[self._open_event_name].trigger)
            self._managers[1].subscribe_off(
                self._events[self._close_event_name].trigger
            )

    def _handle_status(self, db_version, status):
        """Set the status of the dimmable_switch state."""
        self._groups[1].value = status


class NormallyOpenControllerBase(OpenCloseControllerBase):
    """Normally open controller base."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0,
        description="",
        model="",
        state_name=OPEN_CLOSE_SENSOR,
        open_event_name=OPEN_EVENT,
        close_event_name=CLOSE_EVENT,
    ):
        """Init the NormallyOpenControllerBase class."""
        super().__init__(
            address,
            cat,
            subcat,
            firmware=firmware,
            description=description,
            model=model,
            normally_open=True,
            state_name=state_name,
            open_event_name=open_event_name,
            close_event_name=close_event_name,
        )


class NormallyClosedControllerBase(OpenCloseControllerBase):
    """Normally closed controller base."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0,
        description="",
        model="",
        state_name=OPEN_CLOSE_SENSOR,
        open_event_name=OPEN_EVENT,
        close_event_name=CLOSE_EVENT,
    ):
        """Init the NormallyClosedControllerBase class."""
        super().__init__(
            address,
            cat,
            subcat,
            firmware=firmware,
            description=description,
            model=model,
            normally_open=False,
            state_name=state_name,
            open_event_name=open_event_name,
            close_event_name=close_event_name,
        )



"""Handle an outbound direct message to a device."""

import asyncio
from abc import ABCMeta

from .. import ack_handler, direct_ack_handler, direct_nak_handler
from ...address import Address
from ...constants import MessageFlagType
from ..outbound_base import OutboundHandlerBase

TIMEOUT = 3  # Wait time for device response


class DirectCommandHandlerBase(OutboundHandlerBase):
    """Abstract base class for outbound direct message handling."""

    __meta__ = ABCMeta

    def __init__(self, topic, address, group=None):
        """Init the DirectCommandHandlerBase class."""
        self._address = Address(address)
        self._group = group
        self._response_lock = asyncio.Lock()
        super().__init__(
            topic, address=address, group=group, message_type=MessageFlagType.DIRECT
        )

    @property
    def response_lock(self) -> asyncio.Lock:
        """Lock to manage the response between ACK and Direct ACK."""
        return self._response_lock

    async def async_send(self, **kwargs):
        """Send the command and wait for a direct_nak."""
        return await super().async_send(address=self._address, **kwargs)

    # pylint: disable=arguments-differ
    @ack_handler(wait_response=True)
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""

    @direct_nak_handler
    def handle_direct_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the message ACK."""
        pass

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the direct ACK."""



"""Sensor/Actuator devices."""
import asyncio
from ..events import CLOSE_EVENT, OPEN_EVENT, Event, ON_EVENT, OFF_EVENT
from ..extended_property import DELAY, PRESCALER, X10_HOUSE, X10_UNIT
from ..groups import OPEN_CLOSE_SENSOR, RELAY
from ..groups.open_close import NormallyOpen
from ..handlers.to_device.status_request import STATUS_REQUEST, StatusRequestCommand
from ..handlers.to_device.on_level import OnLevelCommand
from ..handlers.to_device.off import OffCommand
from ..managers.on_level_manager import OnLevelManager
from ..operating_flag import (
    LED_BLINK_ON_TX_ON,
    MOMENTARY_FOLLOW_SENSE,
    MOMENTARY_MODE_ON,
    MOMENTARY_ON_OFF_TRIGGER,
    PROGRAM_LOCK_ON,
    RELAY_ON_SENSE_ON,
    X10_OFF,
)
from .commands import ON_COMMAND, OFF_COMMAND
from ..utils import multiple_status
from .on_off_responder_base import OnOffResponderBase
from .device_base import Device
from ..groups.on_off import OnOff
from ..default_link import DefaultLink
from ..constants import RelayMode

ON_LEVEL_MANAGER = "on_level_manager"
RELAY_GROUP = 1
SENSOR_GROUP = 2


class SensorsActuators(OnOffResponderBase):
    """Sensor Actuator device."""


class SensorsActuators_IOLink(Device):
    """I/O Link device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the SensorsActuators_IOLink class."""
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )
        self._op_flags_manager.extended_write = True

    @property
    def relay_mode(self) -> RelayMode:
        """Return the relay mode."""
        if not self._operating_flags[MOMENTARY_MODE_ON].value:
            return RelayMode.LATCHING
        if self._operating_flags[MOMENTARY_FOLLOW_SENSE].value:
            return RelayMode.MOMENTARY_C
        if self._operating_flags[MOMENTARY_ON_OFF_TRIGGER].value:
            return RelayMode.MOMENTARY_B
        return RelayMode.MOMENTARY_A

    async def async_set_relay_mode(self, mode: RelayMode):
        """Set the relay mode and write to device."""
        mode = RelayMode(int(mode))
        self._operating_flags[MOMENTARY_MODE_ON].new_value = not (
            mode == RelayMode.LATCHING
        )
        if mode == RelayMode.MOMENTARY_A:
            self._operating_flags[MOMENTARY_FOLLOW_SENSE].new_value = False
            self._operating_flags[MOMENTARY_ON_OFF_TRIGGER].new_value = False
        if mode == RelayMode.MOMENTARY_B:
            self._operating_flags[MOMENTARY_FOLLOW_SENSE].new_value = False
            self._operating_flags[MOMENTARY_ON_OFF_TRIGGER].new_value = True
        if mode == RelayMode.MOMENTARY_C:
            self._operating_flags[MOMENTARY_FOLLOW_SENSE].new_value = True
            self._operating_flags[MOMENTARY_ON_OFF_TRIGGER].new_value = False
        await self.async_write_op_flags()

    async def async_set_momentary_delay(self, seconds):
        """Set the delay in seconds.

        When the IOLinc is in momentary mode (A, B or C) the relay will
        turn off after the delay period.
        """
        delay = seconds * 10
        prescaler = 1
        if delay > 255:
            prescaler = int(round(delay / 255 + 0.5, 0))
            delay = int(round(delay / prescaler, 0))
        self._properties[PRESCALER].new_value = prescaler
        self._properties[DELAY].new_value = delay
        await self.async_write_ext_properties()

    def on(self, group: int = 0):
        """Turn on the relay."""
        self._handlers[RELAY_GROUP][ON_COMMAND].send()

    async def async_on(self, group: int = 0):
        """Turn on the relay."""
        return await self._handlers[RELAY_GROUP][ON_COMMAND].async_send()

    def off(self, group: int = 0):
        """Turn off the relay."""
        self._handlers[RELAY_GROUP][OFF_COMMAND].send()

    async def async_off(self, group: int = 0):
        """Turn off the relay."""
        return await self._handlers[RELAY_GROUP][OFF_COMMAND].async_send()

    def status(self, group=None):
        """Get the status of the relay and/or the sensor."""
        if group in [1, None]:
            self.relay_status()
        if group == [2, None]:
            self.sensor_status()

    async def async_status(self, group=None):
        """Get the status of the relay and/or the sensor."""
        response_1 = None
        response_2 = None
        if group in [1, None]:
            response_1 = await self.async_relay_status()
        if group in [2, None]:
            response_2 = await self.async_sensor_status()
        return multiple_status(response_1, response_2)

    def relay_status(self):
        """Get the status of the relay switch."""
        self._handlers[RELAY_GROUP][STATUS_REQUEST].send()

    async def async_relay_status(self):
        """Get the status of the relay switch."""
        return await self._handlers[RELAY_GROUP][STATUS_REQUEST].async_send()

    def sensor_status(self):
        """Get the status of the sensor."""
        self._handlers[SENSOR_GROUP][STATUS_REQUEST].send()

    async def async_sensor_status(self):
        """Get the status of the sensor."""
        return await self._handlers[SENSOR_GROUP][STATUS_REQUEST].async_send()

    def _register_operating_flags(self):
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RELAY_ON_SENSE_ON, 0, 2, 4, 5)
        self._add_operating_flag(MOMENTARY_MODE_ON, 0, 3, 6, 7)
        self._add_operating_flag(MOMENTARY_ON_OFF_TRIGGER, 0, 4, 0x12, 0x13)
        self._add_operating_flag(X10_OFF, 0, 5, 0x0C, 0x0D)
        self._add_operating_flag(MOMENTARY_FOLLOW_SENSE, 0, 6, 0x0E, 0x0F)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 7, 0x14, 0x15)

        self._add_property(PRESCALER, 3, 7)
        self._add_property(DELAY, 4, 6)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[RELAY_GROUP] = {}
        self._handlers[RELAY_GROUP][ON_COMMAND] = OnLevelCommand(
            self._address, RELAY_GROUP
        )
        self._handlers[RELAY_GROUP][OFF_COMMAND] = OffCommand(
            self._address, RELAY_GROUP
        )
        self._handlers[RELAY_GROUP][STATUS_REQUEST] = StatusRequestCommand(
            self._address
        )

        # Sensor group controls group 1, not group 2
        self._managers[SENSOR_GROUP] = {}
        self._managers[SENSOR_GROUP][ON_LEVEL_MANAGER] = OnLevelManager(
            self._address, 1
        )
        self._handlers[SENSOR_GROUP] = {}
        self._handlers[SENSOR_GROUP][STATUS_REQUEST] = StatusRequestCommand(
            self.address, 1
        )

    def _register_groups(self):
        self._groups[RELAY_GROUP] = OnOff(RELAY, self._address, RELAY_GROUP)
        self._groups[SENSOR_GROUP] = NormallyOpen(
            OPEN_CLOSE_SENSOR, self._address, SENSOR_GROUP
        )

    def _register_events(self):
        self._events[RELAY_GROUP] = {}
        self._events[RELAY_GROUP][ON_EVENT] = Event(
            ON_EVENT, self._address, RELAY_GROUP
        )
        self._events[RELAY_GROUP][OFF_EVENT] = Event(
            OFF_EVENT, self._address, RELAY_GROUP
        )
        self._events[SENSOR_GROUP] = {}
        self._events[SENSOR_GROUP][OPEN_EVENT] = Event(
            OPEN_EVENT, self._address, SENSOR_GROUP
        )
        self._events[SENSOR_GROUP][CLOSE_EVENT] = Event(
            CLOSE_EVENT, self._address, SENSOR_GROUP
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        switch_on_event = self._events[RELAY_GROUP][ON_EVENT]
        switch_off_event = self._events[RELAY_GROUP][OFF_EVENT]
        on_cmd = self._handlers[RELAY_GROUP][ON_COMMAND]
        off_cmd = self._handlers[RELAY_GROUP][OFF_COMMAND]
        switch_status_cmd = self._handlers[RELAY_GROUP][STATUS_REQUEST]

        on_cmd.subscribe(self._switch_changed)
        off_cmd.subscribe(self._switch_changed)
        on_cmd.subscribe(switch_on_event.trigger)
        off_cmd.subscribe(switch_off_event.trigger)
        switch_status_cmd.subscribe(self._handle_switch_status)

        sensor = self._groups[SENSOR_GROUP]
        close_event = self._events[SENSOR_GROUP][CLOSE_EVENT]
        open_event = self._events[SENSOR_GROUP][OPEN_EVENT]
        manager = self._managers[SENSOR_GROUP][ON_LEVEL_MANAGER]
        sensor_status_cmd = self._handlers[SENSOR_GROUP][STATUS_REQUEST]

        manager.subscribe(sensor.set_value)
        manager.subscribe_on(open_event.trigger)
        manager.subscribe_off(close_event.trigger)
        sensor_status_cmd.subscribe(self._handle_sensor_status)

    def _register_default_links(self):
        link = DefaultLink(
            is_controller=True,
            group=RELAY_GROUP,
            dev_data1=0,
            dev_data2=0,
            dev_data3=0,
            modem_data1=0,
            modem_data2=0,
            modem_data3=0,
        )
        self._default_links.append(link)

    def _switch_changed(self, on_level):
        """Catch on/off signal and fire appropriate response."""
        self._groups[RELAY_GROUP].value = on_level
        if on_level:
            if self._operating_flags[MOMENTARY_MODE_ON].value:
                asyncio.ensure_future(self._delay_wait())

    def _calc_delay(self):
        """Calculate the momentary delay based on properties and flags."""
        delay = self._properties[DELAY].value
        delay = 255 if delay == 0 else delay
        if self._properties[PRESCALER].value is not None:
            prescaler = max(self._properties[PRESCALER].value, 1)
        else:
            prescaler = 1
        return delay * prescaler / 10

    async def _delay_wait(self):
        await asyncio.sleep(self._calc_delay())
        self._groups[RELAY_GROUP].value = 0
        self._events[RELAY_GROUP][OFF_EVENT].trigger(on_level=0)

    def _handle_switch_status(self, db_version, status):
        """Handle status response."""
        self._groups[RELAY_GROUP].set_value(status)

    def _handle_sensor_status(self, db_version, status):
        """Handle status response."""
        self._groups[SENSOR_GROUP].set_value(status)



"""On / Off state."""
from .group_base import GroupBase
from ..address import Address


class OnOff(GroupBase):
    """On / Off state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: int = None
    ):
        """Init the OnLevel class."""
        super().__init__(name, address, group, default, value_type=int)

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        self.value = 0xFF if on_level else 0


class LowBattery(GroupBase):
    """Low battery state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: bool = None
    ):
        """Init the LowBattery class."""
        super().__init__(name, address, group, default, value_type=bool)

    # pylint: disable=arguments-differ
    def set_value(self, low_battery):
        """Set the value of the state from the handlers."""
        self.value = bool(low_battery)


class Heartbeat(GroupBase):
    """Heartbeat state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: bool = None
    ):
        """Init the Hearbeat class."""
        super().__init__(name, address, group, default, value_type=bool)

    # pylint: disable=arguments-differ
    def set_value(self, heartbeat):
        """Set the value of the state from the handlers."""
        self.value = bool(heartbeat)



"""Base device object."""

import asyncio
import logging
from functools import partial
from inspect import getfullargspec

from ..utils import subscribe_topic
from ..aldb.aldb_battery import ALDBBattery
from ..constants import ResponseStatus
from ..handlers.to_device.extended_set import ExtendedSetCommand

_LOGGER = logging.getLogger(__name__)
TIMEOUT = 2


# pylint: disable=no-member
class BatteryDeviceBase:
    """Base class for battery operated devices."""

    def __init__(
        self, address, cat, subcat, firmware=0x00, description="", model="", **kwargs
    ):
        """Init the DeviceBattery class."""

        # The super class is the non-battery base class such as OnOffControllerBase
        super(BatteryDeviceBase, self).__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=0x00,
            description=description,
            model=model,
            **kwargs,
        )
        self._is_battery = True
        self._commands_queue = asyncio.Queue()
        self._aldb = ALDBBattery(address=address, run_command=self._run_on_wake)
        self._last_run = None
        self._keep_awake_cmd = ExtendedSetCommand(self._address, data1=0, data2=0x04)
        subscribe_topic(self._device_awake, self._address.id)
        # This may or may not make sense.
        # It it helps to set default links when the device first connect,
        # then this probably makes sense.
        asyncio.ensure_future(self.async_keep_awake())

    def _run_on_wake(self, command, retries=3, **kwargs):
        cmd = partial(command, **kwargs)
        self._commands_queue.put_nowait((cmd, retries))
        return ResponseStatus.RUN_ON_WAKE

    def close(self):
        """Close the command listener."""
        self._commands_queue.put_nowait((None, None))

    async def async_status(self, group=None):
        """Get device status."""
        args = getfullargspec(super(BatteryDeviceBase, self).async_status)
        if "group" in args[0]:
            return self._run_on_wake(
                super(BatteryDeviceBase, self).async_status, group=group
            )
        return self._run_on_wake(super(BatteryDeviceBase, self).async_status)

    async def async_read_op_flags(self):
        """Read the device operating flags."""
        return self._run_on_wake(super(BatteryDeviceBase, self).async_read_op_flags)

    async def async_write_op_flags(self):
        """Write the operating flags to the device."""
        return self._run_on_wake(super(BatteryDeviceBase, self).async_write_op_flags)

    async def async_read_ext_properties(self):
        """Get the device extended properties."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_read_ext_properties
        )

    async def async_get_engine_version(self):
        """Read the device engine version."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_get_engine_version
        )

    async def async_add_default_links(self):
        """Add default links to the device."""
        return self._run_on_wake(super(BatteryDeviceBase, self).async_add_default_links)

    async def async_read_config(self):
        """Get all configuration settings.

        This includes:
        - Operating flags
        - Extended properties
        - All-Link Database records.
        """
        return self._run_on_wake(super(BatteryDeviceBase, self).async_add_default_links)

    async def async_read_product_id(self):
        """Get the product ID."""
        return self._run_on_wake(super(BatteryDeviceBase, self).async_read_product_id)

    async def async_write_ext_properties(self):
        """Write the extended properties."""
        return self._run_on_wake(
            super(BatteryDeviceBase, self).async_write_ext_properties
        )

    async def async_keep_awake(self, awake_time=0xFF):
        """Keep the device awake to ensure commands are heard."""
        return await self._keep_awake_cmd.async_send(data3=awake_time, priority=1)

    def _device_awake(self, **kwargs):
        """Execute the commands that were requested while sleeping."""
        if self._commands_queue.empty():
            return
        if self._last_run is None or self._last_run.done():
            _LOGGER.debug("We have commands to run so let's get to it")
            self._last_run = asyncio.ensure_future(self._ensure_commands())

    async def _ensure_commands(self):
        """Ensure any commands are run."""
        keep_awake_retry = 3
        while keep_awake_retry:
            result = await self.async_keep_awake()
            _LOGGER.debug("Keep awake result: %s", str(result))
            if result == ResponseStatus.SUCCESS:
                return await self._run_commands()
            keep_awake_retry -= 1
            _LOGGER.debug("Retries: %d", keep_awake_retry)
            await asyncio.sleep(3)

    async def _run_commands(self):
        """Run the commands in queue."""
        retry_cmds = []
        while True:
            try:
                command, retries = await asyncio.wait_for(
                    self._commands_queue.get(), TIMEOUT
                )
                _LOGGER.debug("got a command to run YAY")
                _LOGGER.debug(str(command))
                if command is None:
                    return
                result = await command()
                if result != ResponseStatus.SUCCESS and retries:
                    retry_cmds.append((command, retries - 1))
            except asyncio.TimeoutError:
                break

        for cmd, retries in retry_cmds:
            self._run_on_wake(cmd, retries)



"""Extended Message User Data Type."""
import binascii
import logging

from ...utils import vars_to_string

_LOGGER = logging.getLogger(__name__)


def create_from_dict(user_data_dict: dict):
    """Create a UserData object from a dictionary."""
    empty = create_empty()
    user_data = _dict_to_dict(empty, user_data_dict)
    return user_data


def create_empty(val=0x00):
    """Create an empty Userdata object.

    val: value to fill the empty user data fields with (default is 0x00)
    """
    user_data_dict = {}
    for index in range(1, 15):
        key = "d{}".format(index)
        user_data_dict.update({key: val})
    return user_data_dict


def _dict_to_dict(empty, user_data):
    if isinstance(user_data, dict):
        for key in user_data:
            if key in [
                "d1",
                "d2",
                "d3",
                "d4",
                "d5",
                "d6",
                "d7",
                "d8",
                "d9",
                "d10",
                "d11",
                "d12",
                "d13",
                "d14",
            ]:
                empty[key] = user_data[key]
    return empty


def _bytes_to_dict(empty, user_data):
    if len(user_data) == 14:
        for index in range(1, 15):
            key = "d{}".format(index)
            empty[key] = user_data[index - 1]
    else:
        raise ValueError
    return empty


def _normalize(empty, user_data):
    """Return normalized user data as a dictionary.

    empty: an empty dictionary
    user_data: data in the form of Userdata, dict or None
    """
    if isinstance(user_data, UserData):
        return user_data.to_dict()
    if isinstance(user_data, dict):
        return _dict_to_dict(empty, user_data)
    if isinstance(user_data, (bytes, bytearray)):
        return _bytes_to_dict(empty, user_data)
    if user_data is None:
        return empty
    raise ValueError


class UserData:
    """Extended Message User Data Type."""

    def __init__(self, user_data=bytearray(bytes(14))):
        """Init the Userdata Class."""
        self._user_data = _normalize(create_empty(0x00), user_data)

    def __len__(self):
        """Init Userdata Class."""
        return len(self._user_data)

    def __iter__(self):
        """Iterate through the user data bytes."""
        for itm in self._user_data:
            yield itm

    def __getitem__(self, key):
        """Return a single byte of the user data."""
        return self._user_data.get(key)

    def __setitem__(self, key, val):
        """Set a user data element."""
        self._user_data[key] = val

    def __eq__(self, other):
        """Test if the current user data equals another user data instance."""
        isequal = False
        if isinstance(other, UserData):
            for key in self._user_data:
                if self._user_data[key] == other[key]:
                    isequal = True
                else:
                    isequal = False
                    break
        return isequal

    def __ne__(self, other):
        """Test if the current user data is not equal to another instance."""
        return bool(self != other)

    def __repr__(self):
        """Emit the user data in debug format (AA.BB.CC)."""
        hex_str = binascii.hexlify(bytes(self)).decode()
        strout = ""
        first = True
        for index in range(0, 28, 2):
            if first:
                first = False
            else:
                strout = strout + "."
            strout = strout + hex_str[index : index + 2]
        return strout

    def __str__(self):
        """Emit the user data in human readable format."""
        data = []
        for index in range(1, 15):
            key = "d{}".format(index)
            val = self._user_data[key]
            data.append((key, val))
        return vars_to_string(data)

    def __bytes__(self):
        """Emit the address in bytes format."""
        byteout = bytearray()
        for index in range(1, 15):
            key = "d" + str(index)
            if self._user_data.get(key) is not None:
                byteout.append(self._user_data[key])
            else:
                byteout.append(0x00)
        return bytes(byteout)

    def get(self, key):
        """Return a single byte of the user data."""
        return self._user_data.get(key)

    def to_dict(self):
        """Return user_data as a dict object."""
        return self._user_data

    def set_checksum(self, cmd1: int, cmd2: int, version=2):
        """Set the checksum."""
        data_sum = cmd1 + cmd2
        for index in range(1, 14):
            data_sum += self._user_data["d{:d}".format(index)]
        chksum = 0xFF - (data_sum & 0xFF) + 1
        self._user_data["d14"] = chksum & 0xFF

    def set_crc(self, cmd1: int, cmd2: int):
        """Set Userdata[13] and Userdata[14] to the CRC value."""
        data1 = bytes([cmd1])
        data2 = bytes([cmd2])
        data3 = bytes(self)[0:12]
        data = b"".join([data1, data2, data3])
        crc = int(0)
        for curr_byte in data:
            for _ in range(0, 8):
                fbit = curr_byte & 0x01
                fbit = fbit ^ 0x01 if (crc & 0x8000) else fbit
                fbit = fbit ^ 0x01 if (crc & 0x4000) else fbit
                fbit = fbit ^ 0x01 if (crc & 0x1000) else fbit
                fbit = fbit ^ 0x01 if (crc & 0x0008) else fbit
                crc = ((crc << 1) | fbit) & 0xFFFF
                curr_byte = curr_byte >> 1
        self._user_data["d13"] = (crc >> 8) & 0xFF
        self._user_data["d14"] = crc & 0xFF



"""Dimmable Lighting Control Devices (CATEGORY 0x01)."""
from ..default_link import DefaultLink
from ..events import OFF_EVENT, ON_EVENT, Event
from ..groups import ON_OFF_SWITCH
from ..groups.on_off import OnOff
from ..handlers.to_device.status_request import StatusRequestCommand
from ..managers.on_level_manager import OnLevelManager
from .commands import STATUS_COMMAND, STATUS_COMMAND_HUB
from .device_base import Device

ON_LEVEL_MANAGER = "on_level_manager"


class OnOffControllerBase(Device):
    """Base device for ON/OFF controllers."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
        on_event_name=ON_EVENT,
        off_event_name=OFF_EVENT,
        on_fast_event_name=None,
        off_fast_event_name=None,
    ):
        """Init the OnOffControllerBase class."""

        self._buttons = {1: ON_OFF_SWITCH} if buttons is None else buttons
        self._on_event_name = on_event_name
        self._off_event_name = off_event_name
        self._on_fast_event_name = on_fast_event_name
        self._off_fast_event_name = off_fast_event_name

        super().__init__(address, cat, subcat, firmware, description, model)

    def status(self, group=None):
        """Request the status of the device."""
        self._handlers[STATUS_COMMAND].send()

    async def async_status(self, group=None):
        """Request the status of the device."""
        return await self._handlers[STATUS_COMMAND].async_send()

    def _register_default_links(self):
        """Register default links for the device."""
        super()._register_default_links()
        for group in self._buttons:
            link = DefaultLink(
                is_controller=True,
                group=group,
                dev_data1=255,
                dev_data2=28,
                dev_data3=group,
                modem_data1=0,
                modem_data2=0,
                modem_data3=0,
            )
            self._default_links.append(link)

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()

        self._handlers[STATUS_COMMAND] = StatusRequestCommand(self._address, 0)
        self._handlers[STATUS_COMMAND_HUB] = StatusRequestCommand(self._address, 2)
        for group in self._buttons:
            if self._managers.get(group) is None:
                self._managers[group] = {}
            self._managers[group][ON_LEVEL_MANAGER] = OnLevelManager(
                self._address, group
            )

    def _register_groups(self):
        super()._register_groups()
        for group in self._buttons:
            if self._buttons[group] is not None:
                name = self._buttons[group]
                self._groups[group] = OnOff(name, self._address, group)

    def _register_events(self):
        super()._register_events()
        for group in self._buttons:
            self._events[group] = {}
            button = self._buttons[group]
            if self._on_event_name:
                self._events[group][self._on_event_name] = Event(
                    self._on_event_name, self._address, group, button
                )

            if self._off_event_name:
                self._events[group][self._off_event_name] = Event(
                    self._off_event_name, self._address, group, button
                )

            if self._on_fast_event_name:
                self._events[group][self._on_fast_event_name] = Event(
                    self._on_fast_event_name, self._address, group, button
                )

            if self._off_fast_event_name:
                self._events[group][self._off_fast_event_name] = Event(
                    self._off_fast_event_name, self._address, group, button
                )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[STATUS_COMMAND].subscribe(self._handle_status)
        self._handlers[STATUS_COMMAND_HUB].subscribe(self._handle_status)
        for group in self._buttons:
            if self._groups.get(group) is not None:
                self._managers[group][ON_LEVEL_MANAGER].subscribe(
                    self._groups[group].set_value
                )
            if self._on_event_name:
                event = self._events[group][self._on_event_name]
                self._managers[group][ON_LEVEL_MANAGER].subscribe_on(event.trigger)

            if self._off_event_name:
                event = self._events[group][self._off_event_name]
                self._managers[group][ON_LEVEL_MANAGER].subscribe_off(event.trigger)

            if self._on_fast_event_name:
                event = self._events[group][self._on_fast_event_name]
                self._managers[group][ON_LEVEL_MANAGER].subscribe_on_fast(event.trigger)

            if self._off_fast_event_name:
                event = self._events[group][self._off_fast_event_name]
                self._managers[group][ON_LEVEL_MANAGER].subscribe_off_fast(
                    event.trigger
                )

    def _handle_status(self, db_version, status):
        """Handle status response."""
        # Add this to a separate handler for devices that the cmd1 field
        # returns the ALDB Versioh
        # self.aldb.version = db_version
        self._groups[1].set_value(status)



"""Create outbound messages."""
import logging
from ... import pub
from ...address import Address
from ...constants import (
    AllLinkMode,
    DeviceCategory,
    ManageAllLinkRecordAction,
    MessageId,
    X10CommandType,
)
from ...topics import (
    ALL_LINK_CLEANUP_STATUS_REPORT,
    CANCEL_ALL_LINKING,
    GET_ALL_LINK_RECORD_FOR_SENDER,
    GET_FIRST_ALL_LINK_RECORD,
    GET_IM_CONFIGURATION,
    GET_IM_INFO,
    GET_NEXT_ALL_LINK_RECORD,
    LED_OFF,
    LED_ON,
    MANAGE_ALL_LINK_RECORD,
    RESET_IM,
    RF_SLEEP,
    SEND_ALL_LINK_COMMAND,
    SEND_EXTENDED,
    SEND_STANDARD,
    SET_ACK_MESSAGE_BYTE,
    SET_ACK_MESSAGE_TWO_BYTES,
    SET_HOST_DEV_CAT,
    SET_IM_CONFIGURATION,
    SET_NAK_MESSAGE_BYTE,
    START_ALL_LINKING,
    X10_SEND,
)
from ...utils import publish_topic, subscribe_topic
from ..topic_converters import topic_to_message_handler, topic_to_message_type
from . import MessageBase
from .all_link_record_flags import AllLinkRecordFlags
from .im_config_flags import IMConfigurationFlags
from .message_definition import MessageDefinition
from .message_definitions import FLD_EXT_SEND, FLD_STD_SEND, OUTBOUND_MSG_DEF
from .message_flags import MessageFlags
from .message_flags import create as create_flags
from .user_data import UserData

# pylint: disable=invalid-name
topic_register = {}

_LOGGER = logging.getLogger(__name__)


def register_outbound_handlers():
    """Register outbound handlers."""
    for topic in topic_register:
        func = topic_register[topic]
        subscribe_topic(func, topic)


class Outbound(MessageBase):
    """Outbound message class."""

    def __eq__(self, other):
        """Needed to allow a prioritized queue work.

        Always returns True.
        """
        return True

    def __gt__(self, other):
        """Needed to allow a prioritized queue work.

        Always returns False.
        """
        return False

    def __lt__(self, other):
        """Needed to allow a prioritized queue work.

        Always returns False.
        """
        return False

    def __hash__(self):
        """Needed to allow a prioritized queue work.

        Return hash of str(self).
        """
        return hash(str(self))


def _create_outbound_message(topic, priority=5, **kwargs) -> Outbound:
    """Create an Outbound message."""
    topic = topic.name.split(".")[1]
    msg_id = getattr(MessageId, topic.upper())
    msg_def = OUTBOUND_MSG_DEF[msg_id]
    msg = Outbound(msg_def, **kwargs)
    publish_topic("send_message.{}".format(topic), msg=msg, priority=priority)


@topic_to_message_handler(
    register_list=topic_register, topic=ALL_LINK_CLEANUP_STATUS_REPORT
)
def all_link_cleanup_status_report(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a ALL_LINK_CLEANUP_STATUS_REPORT outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(register_list=topic_register, topic=GET_IM_INFO)
def get_im_info(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_IM_INFO outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(register_list=topic_register, topic=SEND_ALL_LINK_COMMAND)
def send_all_link_command(
    group: int, cmd1: int, cmd2: int, topic=pub.AUTO_TOPIC
) -> Outbound:
    """Create a SEND_ALL_LINK_COMMAND outbound message."""
    _create_outbound_message(group=group, cmd1=cmd1, cmd2=cmd2, topic=topic, priority=3)


def _create_flags(topic, extended):
    msg_type = topic_to_message_type(topic)
    return create_flags(msg_type, extended=extended)


@topic_to_message_handler(register_list=topic_register, topic=SEND_STANDARD)
def send_standard(
    address: Address,
    cmd1: int,
    cmd2: int,
    flags: MessageFlags = None,
    priority=5,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a SEND_STANDARD outbound message."""
    subtopics = topic.name.split(".")
    main_topic = subtopics[1]
    msg_type = None if len(subtopics) < 3 else subtopics[2]
    flags = flags if flags is not None else _create_flags(topic, False)
    kwargs = {"address": address, "flags": flags, "cmd1": cmd1, "cmd2": cmd2}
    msg_def = MessageDefinition(MessageId.SEND_EXTENDED, FLD_STD_SEND)
    send_topic = "send_message.{}".format(main_topic)
    if msg_type is not None:
        send_topic = "{}.{}".format(send_topic, msg_type)
    msg = Outbound(msg_def, **kwargs)
    publish_topic(send_topic, msg=msg, priority=priority)


@topic_to_message_handler(register_list=topic_register, topic=SEND_EXTENDED)
def send_extended(
    address: Address,
    cmd1: int,
    cmd2: int,
    user_data: UserData,
    flags: MessageFlags = None,
    priority=5,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a SEND_EXTENDED outbound message."""
    main_topic = topic.name.split(".")[1]
    flags = flags if flags is not None else _create_flags(topic, True)
    kwargs = {
        "address": address,
        "flags": flags,
        "cmd1": cmd1,
        "cmd2": cmd2,
        "user_data": user_data,
    }
    msg_def = MessageDefinition(MessageId.SEND_EXTENDED, FLD_EXT_SEND)
    send_topic = "send_message.{}".format(main_topic)
    msg = Outbound(msg_def, **kwargs)
    publish_topic(send_topic, msg=msg, priority=priority)


@topic_to_message_handler(register_list=topic_register, topic=X10_SEND)
def x10_send(raw_x10: int, x10_flag: X10CommandType, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a X10_SEND outbound message."""
    _create_outbound_message(raw_x10=raw_x10, x10_flag=x10_flag, topic=topic)


@topic_to_message_handler(register_list=topic_register, topic=START_ALL_LINKING)
def start_all_linking(mode: AllLinkMode, group: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a START_ALL_LINKING outbound message."""
    _create_outbound_message(mode=mode, group=group, topic=topic, priority=7)


@topic_to_message_handler(register_list=topic_register, topic=CANCEL_ALL_LINKING)
def cancel_all_linking(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a CANCEL_ALL_LINKING outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(register_list=topic_register, topic=SET_HOST_DEV_CAT)
def set_host_dev_cat(
    cat: DeviceCategory, subcat: int, firmware: int, topic=pub.AUTO_TOPIC
) -> Outbound:
    """Create a SET_HOST_DEV_CAT outbound message."""
    _create_outbound_message(
        cat=cat, subcat=subcat, firmware=firmware, topic=topic, priority=7
    )


@topic_to_message_handler(register_list=topic_register, topic=RESET_IM)
def reset_im(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a RESET_IM outbound message."""
    _create_outbound_message(topic=topic, priority=2)


@topic_to_message_handler(register_list=topic_register, topic=SET_ACK_MESSAGE_BYTE)
def set_ack_message_byte(cmd2: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a SET_ACK_MESSAGE_BYTE outbound message."""
    _create_outbound_message(cmd2=cmd2, topic=topic, priority=7)


@topic_to_message_handler(register_list=topic_register, topic=GET_FIRST_ALL_LINK_RECORD)
def get_first_all_link_record(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_FIRST_ALL_LINK_RECORD outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(register_list=topic_register, topic=GET_NEXT_ALL_LINK_RECORD)
def get_next_all_link_record(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_NEXT_ALL_LINK_RECORD outbound message."""
    _create_outbound_message(topic=topic, priority=1)


@topic_to_message_handler(register_list=topic_register, topic=SET_IM_CONFIGURATION)
def set_im_configuration(
    disable_auto_linking: bool,
    monitor_mode: bool,
    auto_led: bool,
    deadman: bool,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a SET_IM_CONFIGURATION outbound message."""
    flag_byte = 0x00
    flag_byte = flag_byte | (1 if disable_auto_linking else 0) << 7
    flag_byte = flag_byte | (1 if monitor_mode else 0) << 6
    flag_byte = flag_byte | (1 if auto_led else 0) << 5
    flag_byte = flag_byte | (1 if deadman else 0) << 4
    flags = IMConfigurationFlags(flag_byte)
    _create_outbound_message(flags=flags, topic=topic, priority=2)


@topic_to_message_handler(
    register_list=topic_register, topic=GET_ALL_LINK_RECORD_FOR_SENDER
)
def get_all_link_record_for_sender(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_ALL_LINK_RECORD_FOR_SENDER outbound message."""
    _create_outbound_message(topic=topic, priority=2)


@topic_to_message_handler(register_list=topic_register, topic=LED_ON)
def led_on(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a LED_ON outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(register_list=topic_register, topic=LED_OFF)
def led_off(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a LED_OFF outbound message."""
    _create_outbound_message(topic=topic, priority=7)


@topic_to_message_handler(register_list=topic_register, topic=MANAGE_ALL_LINK_RECORD)
def manage_all_link_record(
    action: ManageAllLinkRecordAction,
    flags: AllLinkRecordFlags,
    group: int,
    target: Address,
    data1: int,
    data2: int,
    data3: int,
    topic=pub.AUTO_TOPIC,
) -> Outbound:
    """Create a MANAGE_ALL_LINK_RECORD outbound message."""
    _create_outbound_message(
        action=action,
        flags=flags,
        group=group,
        target=target,
        data1=data1,
        data2=data2,
        data3=data3,
        topic=topic,
        priority=10,
    )


@topic_to_message_handler(register_list=topic_register, topic=SET_NAK_MESSAGE_BYTE)
def set_nak_message_byte(cmd2: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a SET_NAK_MESSAGE_BYTE outbound message."""
    _create_outbound_message(cmd2=cmd2, topic=topic, priority=10)


@topic_to_message_handler(register_list=topic_register, topic=SET_ACK_MESSAGE_TWO_BYTES)
def set_ack_message_two_bytes(cmd1: int, cmd2: int, topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a SET_ACK_MESSAGE_TWO_BYTES outbound message."""
    _create_outbound_message(cmd1=cmd1, cmd2=cmd2, topic=topic, priority=10)


@topic_to_message_handler(register_list=topic_register, topic=RF_SLEEP)
def rf_sleep(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a RF_SLEEP outbound message."""
    _create_outbound_message(topic=topic, priority=2)


@topic_to_message_handler(register_list=topic_register, topic=GET_IM_CONFIGURATION)
def get_im_configuration(topic=pub.AUTO_TOPIC) -> Outbound:
    """Create a GET_IM_CONFIGURATION outbound message."""
    _create_outbound_message(topic=topic, priority=7)
